"""Automated tests for the analysis functions."""

import unittest
import os
import copy
import logging
from unittest.mock import patch
import numpy as np
from numpy.testing import assert_allclose
import importlib
import pyspedas
from pyspedas import (
    avg_data,
    yclip,
    deriv_data,
    tinterpol,
    tvectot,
    wavelet,
    time_domain_filter,
    get_data,
    store_data,
    replace_data,
    time_string,
    time_float,
    data_exists,
    del_data,
    tcrossp,
    tdotp,
    tnormalize,
    smooth,
    subtract_average,
    subtract_median,
    tsmooth,
    time_clip,
    tdeflag,
    clean_spikes,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file


# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
analysis_dir = "analysis_tools"
save_dir = os.path.join(output_dir, analysis_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class TestAnalysisFunctions(unittest.TestCase):
    """Test functions under analysis folder."""

    def setUp(self):
        """Create a tplot variable to be used in tests."""
        del_data("*")
        store_data(
            "test",
            data={
                "x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                "y": [3.0, 5.0, 8.0, 15.0, 20.0, 1.0],
            },
        )

    def tearDown(self):
        """Delete the tplot variable."""
        del_data("*")

    def test_tdotp(self):
        store_data("var1", data={"x": [0], "y": [[3, -3, 1]]})
        store_data("var2", data={"x": [0], "y": [[4, 9, 2]]})
        _ = tdotp("var1", "var2")
        dpdata = get_data("var1_dot_var2")
        self.assertTrue(dpdata.y == np.array([-13]))

    def test_tcrossp(self):
        """cross product tests"""
        cp = tcrossp([3, -3, 1], [4, 9, 2], return_data=True)
        self.assertTrue(cp.tolist() == [-15, -2, 39])
        cp = tcrossp([3, -3, 1], [4, 9, 2])
        cp = get_data(cp)
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])

        store_data("var1", data={"x": [0], "y": [[3, -3, 1]]})
        store_data("var2", data={"x": [0], "y": [[4, 9, 2]]})
        cp = tcrossp("var1", "var2", return_data=True)
        self.assertTrue(cp[0].tolist() == [-15, -2, 39])

        cp = tcrossp("var1", "var2", newname="test_crossp")
        cp = get_data("test_crossp")
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])

    def test_tnormalize(self):
        """tests for normalizing tplot variables"""
        store_data(
            "test_tnormalize",
            data={
                "x": [1, 2, 3, 4, 5],
                "y": [[3, 2, 1], [1, 2, 3], [10, 5, 1], [8, 10, 14], [70, 20, 10]],
            },
        )
        norm = tnormalize("test_tnormalize")
        normalized_data = get_data(norm)
        self.assertTrue(
            np.round(normalized_data.y[0, :], 4).tolist() == [0.8018, 0.5345, 0.2673]
        )
        self.assertTrue(
            np.round(normalized_data.y[1, :], 4).tolist() == [0.2673, 0.5345, 0.8018]
        )
        self.assertTrue(
            np.round(normalized_data.y[2, :], 4).tolist() == [0.8909, 0.4454, 0.0891]
        )
        self.assertTrue(
            np.round(normalized_data.y[3, :], 4).tolist() == [0.4216, 0.527, 0.7379]
        )
        self.assertTrue(
            np.round(normalized_data.y[4, :], 4).tolist() == [0.9526, 0.2722, 0.1361]
        )

    def test_subtract_median(self):
        """Test subtract_median."""

        none1 = subtract_median("aaabbbccc")  # Test non-existent name
        self.assertTrue(none1 is None)
        subtract_median("test")
        d = get_data("test-m")
        self.assertTrue(d[1].tolist() == [-3.5, -1.5, 1.5, 8.5, 13.5, -5.5])
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, float("nan")],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        none2 = subtract_median("aaabbbcc")
        self.assertTrue(none2 is None)
        subtract_median("test1", newname="aabb")
        d = get_data("aabb")
        self.assertTrue(len(d[1]) == 6)
        subtract_median(["test", "aabb"], newname="aaabbb")
        subtract_median("test1", overwrite=True)
        subtract_average("test", newname="testtest")
        subtract_average(["test-m", "test"], newname="testtest2")

    def test_subtract_average(self):
        """Test subtract_average."""
        subtract_average("aaabbbccc")  # Test non-existent name
        subtract_average("test")
        d = get_data("test-d")
        self.assertTrue(
            (np.round(d[1].tolist()) == [-6.0, -4.0, -1.0, 6.0, 11.0, -8.0]).all()
        )
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, float("nan")],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        subtract_average("aaabbbcc")
        subtract_average("test1", newname="aabb")
        d = get_data("aabb")
        subtract_average(["test", "aabb"], newname="aaabbb")
        subtract_average("test1", overwrite=True)
        subtract_average("test1", newname="testtest")
        subtract_average(["test1", "test"], newname="testtest2")
        self.assertTrue(len(d[1]) == 6)

    def test_subtract_average_nan(self):
        """Test subtract_average with NaN values."""
        # Create a tplot variable with NaN values
        store_data(
            "test-nan", data={"x": [1.0, 2.0, 3.0, 4.0], "y": [3.0, 5.0, np.nan, 8.0]}
        )

        # Run subtract_average
        tvar = subtract_average("test-nan")
        d = get_data(tvar[0])

        # Check that the result is not all NaN
        self.assertFalse(np.isnan(d[1]).all(), "The result should not be all NaN.")

        # Check that NaN remains where it was
        self.assertTrue(np.isnan(d[1][2]), "NaN values should remain in the output.")

        # Check that non-NaN values are properly adjusted
        expected_values = np.array([3.0, 5.0, np.nan, 8.0]) - np.nanmean(
            np.array([3.0, 5.0, np.nan, 8.0])
        )
        np.testing.assert_array_almost_equal(
            d[1],
            expected_values,
            err_msg="Non-NaN values should be adjusted by subtracting the mean.",
        )

        # Additional NaN scenarios
        store_data(
            "test-nan-all",
            data={"x": [1.0, 2.0, 3.0, 4.0], "y": [np.nan, np.nan, np.nan, np.nan]},
        )

        # Check all NaN case
        result_all = subtract_average("test-nan-all")
        d_all = get_data(result_all[0])
        self.assertTrue(
            np.isnan(d_all[1]).all(), "All NaN input should result in all NaN output."
        )

    def test_subtract_median_nan(self):
        """Test subtract_median with NaN values."""
        # Create a tplot variable with NaN values
        store_data(
            "test-nan", data={"x": [1.0, 2.0, 3.0, 4.0], "y": [3.0, 5.0, np.nan, 8.0]}
        )

        # Run subtract_median
        tvar = subtract_median("test-nan")
        d = get_data(tvar[0])

        # Check that the result is not all NaN
        self.assertFalse(np.isnan(d[1]).all(), "The result should not be all NaN.")

        # Check that NaN remains where it was
        self.assertTrue(np.isnan(d[1][2]), "NaN values should remain in the output.")

        # Check that non-NaN values are properly adjusted by subtracting the median
        expected_values = np.array([3.0, 5.0, np.nan, 8.0]) - np.nanmedian(
            np.array([3.0, 5.0, np.nan, 8.0])
        )
        np.testing.assert_array_almost_equal(
            d[1],
            expected_values,
            err_msg="Non-NaN values should be adjusted by subtracting the median.",
        )

        # Additional NaN scenarios
        store_data(
            "test-nan-all",
            data={"x": [1.0, 2.0, 3.0, 4.0], "y": [np.nan, np.nan, np.nan, np.nan]},
        )

        # Check all NaN case
        result_all = subtract_median("test-nan-all")
        d_all = get_data(result_all[0])
        self.assertTrue(
            np.isnan(d_all[1]).all(), "All NaN input should result in all NaN output."
        )

    def test01_subtract_average_with_suffix(self):
        """Test subtract_average with suffix."""
        subtract_average("test", suffix="-sfx")
        self.assertIsNotNone(get_data("test-sfx"))

    def test02_subtract_average_with_newname_and_suffix(self):
        """Test subtract_average with both custom suffix and newname."""
        subtract_average("test", newname="new_test", suffix="-sfx")
        self.assertIsNotNone(get_data("new_test"))
        # Ensure suffix is not applied when newname is provided
        self.assertIsNone(get_data("test-sfx"))

    def test_subtract_average_with_invalid_names(self):
        """Test subtract_average with invalid names input."""
        result = subtract_average("invalid_name")
        self.assertIsNone(result)  # Should handle gracefully

    def test_subtract_average_with_empty_names(self):
        """Test subtract_average with empty names list."""
        result = subtract_average([])
        self.assertIsNone(result)  # Should handle gracefully

    def test_subtract_average_all_same_values(self):
        """Test subtract_average with all same values in the dataset."""
        store_data("test_same", data={"x": [1.0, 2.0, 3.0], "y": [5.0, 5.0, 5.0]})
        subtract_average("test_same")
        d = get_data("test_same-d")
        self.assertTrue((d[1] == [0, 0, 0]).all())

    def test_subtract_average_one_element(self):
        """Test subtract_average with dataset containing only one element."""
        store_data("test_one", data={"x": [1.0], "y": [10.0]})
        subtract_average("test_one")
        d = get_data("test_one-d")
        self.assertTrue((d[1] == [0]).all())

    def test_subtract_median_basic(self):
        """Test basic functionality of subtract_median."""
        store_data("test", data={"x": [1.0, 2.0, 3.0], "y": [4.0, 5.0, 6.0]})
        subtract_median("test")
        d = get_data("test-m")
        self.assertIsNotNone(d)

    def test_subtract_median_parameter_passing(self):
        """Test that parameters are correctly passed to subtract_average via subtract_median."""
        sm = importlib.import_module("pyspedas.tplot_tools.tplot_math.subtract_median")

        with patch.object(sm, "subtract_average") as mock_subtract_average:
            logging.info(mock_subtract_average)
            pyspedas.subtract_median(
                "test", newname="new_test", suffix="-sfx", overwrite=True
            )

            # Check that subtract_average was called with the correct parameters, including median=1
            mock_subtract_average.assert_called_once_with(
                "test", newname="new_test", suffix="-sfx", overwrite=True, median=1
            )

    def test_yclip(self):
        """Test yclip."""
        yclip("aaabbbccc", 0.0, 12.0)  # Test non-existent name
        yclip("aabb", 0.0, 12.0)
        yclip("test", 0.0, 12.0)
        d = get_data("test-clip")
        # Replace nan with -99.0
        dd = np.nan_to_num(d[1], nan=-99.0)
        yclip("test", 0.0, 12.0, newname="name-clip")
        yclip(["test", "name-clip"], 0.0, 12.0, newname="name1-clip")
        yclip("test", 0.0, 12.0, overwrite=True)
        yclip("test", 0.0, 12.0, newname="testtest")
        yclip(["test", "test-clip"], 0.0, 12.0, newname="testtest2")
        self.assertTrue((dd == [3.0, 5.0, 8.0, -99.0, -99.0, 1.0]).all())

    def test_timeclip(self):
        """Test time_clip."""
        time_clip("aaabbbccc", 1577308800, 1577598800)  # Test non-existent
        tn = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800, 1587998800]
        dn = [3.0, 5.0, 8.0, 15.0, 20.0, 1.0]
        store_data("test1", data={"x": tn, "y": dn})
        time_clip("aaabbb", 1577308800, 1577598800)
        time_clip("test1", 1577112800, 1577608800)
        d = get_data("test1-tclip")
        dd = d[1]
        time_clip("test", 1577308800, 1577598800, newname="name-clip")
        time_clip(["test", "name-clip"], 1577308800, 1577598800, newname="name1-ci")
        time_clip("test", 1577308800, 1577598800, overwrite=True)
        time_clip("test", 1577308800, 1577598800, newname="testtest")
        time_clip(["test", "test1"], 1577308800, 1577598800, newname="testtest2")
        time_clip("test1", 1677112800, 1577608800)
        self.assertTrue((dd == [3.0, 5.0, 8.0, 15.0]).all())

    def test_avg_data(self):
        """Test avg_data."""
        avg_data("aaabbbccc", width=2)  # Test non-existent name
        avg_data("test", width=2)
        d = get_data("test-avg")
        self.assertTrue((d[1] == [4.0, 11.5, 10.5]).all())
        avg_data("test", width=2, overwrite=True)  # Test overwrite
        store_data(
            "test",
            data={
                "x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                "y": [3.0, 5.0, 8.0, -4.0, 20.0, 1.0],
            },
        )
        avg_data("test", width=2, newname="aabb")  # Test newname
        d = get_data("aabb")
        # Test multiple names
        avg_data(["test", "aabb"], newname="aaabbb", width=2)
        dt = [1.0, 12.0, 13.0, 14.0, 15.0, 16.0]
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, 1.0],
        ]
        dv = dn
        store_data("test1", data={"x": dt, "y": dn, "v": dv})
        avg_data("test1", width=2)  # Test 3-d data
        avg_data("test1", newname="test2", res=2.0)  # Test a reasonable resolution
        avg_data("test1", res=-1.0)  # Test res error
        avg_data("test1", res=1.0e8)  # Test res error
        d2 = get_data("test2")
        self.assertTrue(len(d2) > 0)
        self.assertTrue(d2[1][-1][0] == 15.0)
        self.assertTrue(len(d2[2]) == len(d2[0]))

    def test_avg_data_idl(self):
        # Compare data with IDL avg_data
        # Requires file: avg_data_validate.tplot from
        #   https://github.com/spedas/test_data
        # The SPEDAS script that creates the file: general/tools/python_validate/avg_data_python_validate.pro

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "avg_data_validate.tplot", save_dir
        )

        trange = ["2010-02-13 00:00:00", "2010-02-13 11:59:59"]
        probe = "b"
        pyspedas.projects.themis.esa(trange=trange, probe=probe)
        var = "thb_peir_en_eflux"
        vara = var + "-avg"
        pyspedas.time_clip(var, trange[0], trange[1], overwrite=True)
        pyspedas.avg_data(var)

        # Get IDL data
        pyspedas.tplot_restore(filename)
        varidl = var + "_avg"

        # Compare IDL to python results
        d1 = pyspedas.get_data(varidl)
        d2 = pyspedas.get_data(vara)

        assert_allclose(d1[0], d2[0], atol=1e-5, rtol=1e-5, equal_nan=True)
        assert_allclose(d1[1], d2[1], atol=1e-5, rtol=1e-5, equal_nan=True)
        assert_allclose(d1[2], d2[2], atol=1e-5, rtol=1e-5, equal_nan=True)

    def test_clean_spikes(self):
        """Test clean_spikes."""
        clean_spikes("aaabbbccc", nsmooth=3)  # Test non-existent name
        clean_spikes("test", nsmooth=3)
        d = get_data("test-despike")
        self.assertTrue(len(d[1]) == 6)
        # Now test 3 dim data.
        dn = [
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [3.0, 5.0, 8.0],
            [15.0, 20.0, 1.0],
            [23.0, 15.0, 28.0],
            [15.0, 20.0, 1.0],
        ]
        store_data("test1", data={"x": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], "y": dn})
        clean_spikes("test1", nsmooth=3)
        d2 = get_data("test1-despike")
        clean_spikes("test", newname="test_desp", nsmooth=3, sub_avg=True)
        clean_spikes(["test", "test1"], newname="test1-desp")
        clean_spikes("test1", overwrite=True)
        self.assertTrue(len(d2[1]) == 6)

    def test_tdeflag(self):
        """Test tdeflag."""
        tdeflag("aaabbbccc")  # Test non-existent name
        dn = [3.0, float("NaN"), 8.0, float("NaN"), 20.0, 1.0]
        len_dn = len(dn)
        replace_data("test", dn)
        tdeflag("test")
        d = get_data("test-deflag")
        tdeflag("test", overwrite=True)
        tdeflag("test", newname="testtest")
        tdeflag(["test", "test-deflag"], newname="testtest2")
        # Length should be two less, because NaNs were removed.
        self.assertTrue(len(d[1]) == len_dn - 2)

    def test_tdeflag_2d(self):
        """Test tdeflag."""
        times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        vals = np.array(
            [
                [1.0, 2.0, 3.0],
                [1.1, 2.1, 3.1],
                [1.2, 2.2, 3.2],
                [1.3, 2.3, 3.3],
                [1.4, 2.4, 3.4],
            ]
        )
        # dn = [3.0, float("NaN"), 8.0, float("NaN"), 20.0, 1.0]

        vals_noflag = copy.deepcopy(vals)
        store_data("vals_noflag", data={"x": times, "y": vals_noflag})
        tdeflag(
            "vals_noflag", newname="vals_noflag_remove"
        )  # no method should default to remove_nan
        tdeflag("vals_noflag", newname="vals_noflag_linear", method="linear")
        tdeflag("vals_noflag", newname="vals_noflag_repeat", method="repeat")
        tdeflag(
            "vals_noflag",
            newname="vals_noflag_replace",
            method="replace",
            fillval=100.0,
        )
        d1 = get_data("vals_noflag_remove")
        d2 = get_data("vals_noflag_linear")
        d3 = get_data("vals_noflag_repeat")
        d4 = get_data("vals_noflag_replace")
        d5 = get_data("vals_noflag_linear")
        np.testing.assert_array_equal(d1.y, vals_noflag)
        np.testing.assert_array_equal(d2.y, vals_noflag)
        np.testing.assert_array_equal(d3.y, vals_noflag)
        np.testing.assert_array_equal(d4.y, vals_noflag)
        np.testing.assert_array_equal(d5.y, vals_noflag)

        # There can be multiple flags. Make sure the 'no unflagged data' case is
        # properly defected when each individual flag leaves some valid data, but
        # the combination flags everything.

        # commented out for now until they pass
        # tdeflag('vals_noflag', newname='vals_all_multiflag_linear', method='linear', flag=[2.0,2.1,2.3,2.4,2.5])
        # self.assertFalse(data_exists('vals_all_multiflag_linear'))
        # tdeflag('vals_noflag', newname='vals_all_multiflag_repeat', method='repeat', flag=[2.0,2.1,2.3,2.4,2.5])
        # self.assertFalse(data_exists('vals_all_multiflag_repeat'))

        vals_flag_beginning = copy.deepcopy(vals)
        vals_flag_beginning[0, 1] = np.nan
        store_data("vals_flag_beginning", data={"x": times, "y": vals_flag_beginning})
        tdeflag(
            "vals_flag_beginning",
            newname="vals_flag_beginning_remove",
            method="remove_nan",
        )
        tdeflag(
            "vals_flag_beginning", newname="vals_flag_beginning_linear", method="linear"
        )
        tdeflag(
            "vals_flag_beginning", newname="vals_flag_beginning_repeat", method="repeat"
        )
        tdeflag(
            "vals_flag_beginning",
            newname="vals_flag_beginning_replace",
            method="replace",
            fillval=100.0,
        )
        d1 = get_data("vals_flag_beginning_remove")
        d2 = get_data("vals_flag_beginning_linear")
        d3 = get_data("vals_flag_beginning_repeat")
        d4 = get_data("vals_flag_beginning_replace")
        np.testing.assert_equal(len(d1.times), 4)
        np.testing.assert_equal(d1.times[0], times[1])
        np.testing.assert_equal(d2.y[0, 1], 2.1)
        np.testing.assert_equal(d3.y[0, 1], 2.1)
        np.testing.assert_equal(d4.y[0, 1], 100.0)

        vals_flag_end = copy.deepcopy(vals)
        vals_flag_end[4, 1] = np.nan
        store_data("vals_flag_end", data={"x": times, "y": vals_flag_end})
        tdeflag("vals_flag_end", newname="vals_flag_end_remove", method="remove_nan")
        tdeflag("vals_flag_end", newname="vals_flag_end_linear", method="linear")
        tdeflag("vals_flag_end", newname="vals_flag_end_repeat", method="repeat")
        tdeflag(
            "vals_flag_end",
            newname="vals_flag_end_replace",
            method="replace",
            fillval=100.0,
        )
        d1 = get_data("vals_flag_end_remove")
        d2 = get_data("vals_flag_end_linear")
        d3 = get_data("vals_flag_end_repeat")
        d4 = get_data("vals_flag_end_replace")
        np.testing.assert_equal(len(d1.times), 4)
        np.testing.assert_equal(d1.times[3], times[3])
        np.testing.assert_equal(d2.y[4, 1], 2.3)
        np.testing.assert_equal(d3.y[4, 1], 2.3)
        np.testing.assert_equal(d4.y[4, 1], 100.0)

        vals_flag_middle = copy.deepcopy(vals)
        vals_flag_middle[2, 1] = np.nan
        store_data("vals_flag_middle", data={"x": times, "y": vals_flag_middle})
        tdeflag(
            "vals_flag_middle", newname="vals_flag_middle_remove", method="remove_nan"
        )
        tdeflag("vals_flag_middle", newname="vals_flag_middle_linear", method="linear")
        tdeflag("vals_flag_middle", newname="vals_flag_middle_repeat", method="repeat")
        tdeflag(
            "vals_flag_middle",
            newname="vals_flag_middle_replace",
            method="replace",
            fillval=100.0,
        )
        d1 = get_data("vals_flag_middle_remove")
        d2 = get_data("vals_flag_middle_linear")
        d3 = get_data("vals_flag_middle_repeat")
        d4 = get_data("vals_flag_middle_replace")
        np.testing.assert_equal(len(d1.times), 4)
        np.testing.assert_equal(d1.times[2], times[3])
        np.testing.assert_equal(d2.y[2, 1], 2.2)
        np.testing.assert_equal(d3.y[2, 1], 2.1)
        np.testing.assert_equal(d4.y[2, 1], 100.0)

        vals_flag_all = copy.deepcopy(vals)
        vals_flag_all[:, :] = np.nan
        store_data("vals_flag_all", data={"x": times, "y": vals_flag_all})
        tdeflag("vals_flag_all", newname="vals_flag_all_remove", method="remove_nan")
        self.assertFalse(data_exists("vals_flag_all_remove"))
        tdeflag("vals_flag_all", newname="vals_flag_all_linear", method="linear")
        self.assertFalse(data_exists("vals_flag_all_linear"))
        tdeflag("vals_flag_all", newname="vals_flag_all_repeat", method="repeat")
        self.assertFalse(data_exists("vals_flag_all_repeat"))
        tdeflag(
            "vals_flag_all",
            newname="vals_flag_all_replace",
            method="replace",
            fillval=100.0,
        )
        d = get_data("vals_flag_all_replace")
        self.assertEqual(d.y[0, 0], 100.0)

    def test_tdeflag_1d(self):
        times = [1.0, 2.0, 3.0, 4.0, 5.0]
        data = [1.0, 2.0, 3.0, 4.0, 5.0]
        vals_noflag = copy.deepcopy(data)
        store_data("vals_noflag_1d", data={"x": times, "y": vals_noflag})
        tdeflag("vals_noflag_1d", newname="vals_noflag_1d_linear", method="linear")
        d1 = get_data("vals_noflag_1d_linear")
        np.testing.assert_array_equal(d1.y, vals_noflag)

    def test_deriv_data(self):
        """Test deriv_data."""
        deriv_data("aaabbbccc")  # Test non-existent name
        deriv_data("test")
        d = get_data("test-der")
        deriv_data("test", overwrite=True)
        deriv_data("test", newname="testtest")
        deriv_data(["test", "test-der"], newname="testtest2")
        self.assertTrue((d[1] == [2.0, 2.5, 5.0, 6.0, -7.0, -19.0]).all())

    def test_tvectot(self):
        from pyspedas.projects.themis import state
        from pyspedas.tplot_tools import data_exists

        state(probe="a")
        tvectot("tha_pos", join_component=True)
        self.assertTrue(data_exists("tha_pos_tot"))
        d = get_data("tha_pos_tot")
        s = d.y.shape[1]
        self.assertEqual(s, 4)
        del_data("tha_pos_tot")
        tvectot("tha_pos")  # join_component defaults to True now
        self.assertTrue(data_exists("tha_pos_tot"))
        d = get_data("tha_pos_tot")
        s = d.y.shape[1]
        self.assertEqual(s, 4)
        del_data("tha_pos_tot")
        _ = tvectot("tha_pos", join_component=False)
        self.assertTrue(data_exists("tha_pos_mag"))
        d = get_data("tha_pos_mag")
        ndims = len(d.y.shape)
        self.assertEqual(ndims, 1)
        del_data("tha_pos_mag")
        _ = tvectot("tha_pos", newname="tha_pos_total", join_component=False)
        self.assertTrue(data_exists("tha_pos_total"))
        d = get_data("tha_pos_total")
        ndims = len(d.y.shape)
        self.assertEqual(ndims, 1)
        del_data("tha_pos_total")
        tvectot("tha_pos", suffix="_rtot")
        self.assertTrue(data_exists("tha_pos_rtot"))

    def test_wavelet(self):
        # Create a tplot variable that contains a wave.
        t = np.arange(4000.0)
        y = np.sin(2 * np.pi * t / 32.0)
        y2 = np.sin(2 * np.pi * t / 64.0)
        y[1000:3000] = y2[1000:3000]
        var = "sin_wav"
        time = time_float("2010-01-01") + 10 * t
        store_data(var, data={"x": time, "y": y})

        # Gaussian Derivative wavelets transformation.
        powervar = wavelet(var, wavename="gaus1")
        pvar = powervar[0]
        self.assertTrue(data_exists(pvar))

    def test_time_domain_filter(self):
        # Create a tplot variable that contains a wave.
        t = np.arange(4000.0)
        y = np.sin(2 * np.pi * t / 32.0)
        y2 = np.sin(2 * np.pi * t / 64.0)
        y[1000:3000] = y2[1000:3000]
        dat = np.zeros((4000, 3))
        dat[:, 0] = y
        dat[:, 1] = y
        dat[:, 2] = y
        time = time_float("2010-01-01") + 10 * t
        output = time_domain_filter(dat, time, 16.0, 48.0)
        self.assertEqual(output.shape[0], dat.shape[0])
        self.assertEqual(output.shape[1], dat.shape[1])

    def test_tsmooth(self):
        """Test smooth."""
        tsmooth("aaabbbccc")  # Test non-existent name
        a = [1.0, 1.0, 2.0, 3.0, 4.0, 1.0, 4.0, 3.0, 2.0, 1.0, 1.0]
        x = smooth(a, 3)
        r = [
            1.0,
            1.3333333333333333,
            2.0,
            3.0,
            2.6666666666666665,
            3.0,
            2.6666666666666665,
            3.0,
            2.0,
            1.3333333333333333,
            1.0,
        ]
        self.assertTrue(x == r)
        b = [1.0, 1.0, 2.0, 3.0, np.nan, np.nan, np.nan, np.nan, 2.0, 1.0, 1.0]
        y = smooth(b, width=3)
        ry = [
            1.0,
            1.3333333333333333,
            2.0,
            1.6666666666666665,
            1.0,
            np.nan,
            np.nan,
            0.6666666666666666,
            1.0,
            1.3333333333333333,
            1.0,
        ]
        self.assertTrue(y == ry)
        tsmooth("test")
        d = get_data("test-s")
        tsmooth("test", overwrite=True)
        tsmooth("test", newname="testtest")
        tsmooth(["test", "test-s"], newname="testtest2")
        self.assertTrue(d[1].tolist() == [3.0, 5.0, 8.0, 15.0, 20.0, 1.0])

    def test_tinterpol(self):
        """Test tinterpol."""
        tinterpol("aaabbbccc", "test")  # Test non-existent name
        tn = [1.0, 1.5, 4.6, 5.8, 6.0]
        dn = [10.0, 15.0, 46.0, 58.0, 60.0]
        data = np.array(
            [
                [0, 1, 2, 3, 4],
                [5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14],
                [15, 16, 17, 18, 19],
                [20, 21, 22, 23, 24],
            ]
        )
        store_data("test1", data={"x": tn, "y": dn})
        store_data("test2", data={"x": tn, "y": data, "v": [10, 20, 30, 40, 50]})
        tinterpol("test1", "test")
        tinterpol("test1", "doesnt_exist")
        tinterpol("test2", "test", newname="")
        tinterpol("test2", [1, 2, 3, 4, 5, 6], newname="pyarray_float")
        tinterpol("test2", np.array([1, 2, 3, 4, 5, 6]), newname="nparray_float")
        tinterpol("test2", time_string([1, 2, 3, 4, 5, 6]), newname="pyarray_str")
        tinterpol(
            "test2", np.array(time_string([1, 2, 3, 4, 5, 6])), newname="nparray_str"
        )
        bad_datatype = {}
        tinterpol("test2", [bad_datatype], newname="nparray_bad_datatype")

        d = get_data("test1-itrp")
        self.assertTrue(d[1][1] == 20.0)
        d2 = get_data("pyarray_float")
        self.assertTrue(abs(d2[1][1][0] - 5.80645161) < 1e-6)
        d3 = get_data("nparray_float")
        self.assertTrue(abs(d3[1][1][0] - 5.80645161) < 1e-6)
        d4 = get_data("pyarray_str")
        self.assertTrue(abs(d4[1][1][0] - 5.80645161) < 1e-6)
        d5 = get_data("nparray_str")
        self.assertTrue(abs(d5[1][1][0] - 5.80645161) < 1e-6)

    def test_scipy_interp1d(self):
        import scipy
        import numpy as np

        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        interpolator = scipy.interpolate.interp1d(
            input_times_float64, values_input, kind="linear"
        )
        result = interpolator(interp_to_time_float64)
        logging.info(result)
        # Known to fail.  This affects xarray.interp and the current version of tinterpol.
        # self.assertTrue((result >= 0.0).all())

    def test_scipy_spline(self):
        import scipy
        import numpy as np

        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        interpolator = scipy.interpolate.make_interp_spline(
            input_times_float64, values_input, k=1
        )
        result = interpolator(interp_to_time_float64)
        logging.info(result)
        # make_interp_spline() with k=1 gives the expected result
        self.assertTrue((result >= 0.0).all())

    def test_xarray_interp(self):
        import xarray as xr
        import numpy as np

        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        data_array = xr.DataArray(
            values_input, dims=["time"], coords={"time": ("time", input_times_npdt64)}
        )

        result = data_array.interp({"time": interp_to_times_npdt64}, method="linear")
        # This is known to fail, due to issues in scipy.interpolate.interp1d
        # logging.info(result.values)
        # result.values is [-3.469446951953614e-18]
        # self.assertTrue((result.values >= 0.0).all())
        self.assertAlmostEqual(result.values, 0.0)

    def test_xarray_interp_float_times(self):
        import xarray as xr
        import numpy as np

        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)

        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(input_times_npdt64[1])

        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)

        data_array = xr.DataArray(
            values_input, dims=["time"], coords={"time": ("time", input_times_float64)}
        )

        result = data_array.interp({"time": interp_to_time_float64}, method="linear")
        # This is known to fail, due to issues in scipy.interpolate.interp1d
        # l ogging.info(result.values)
        # result.values is [-3.469446951953614e-18]
        # self.assertTrue((result.values >= 0.0).all())
        self.assertAlmostEqual(result.values, 0.0)

    def test_numpy_interp(self):
        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)

        time_strings_interp_to = np.array(["2018-07-01T13:02:16.922475008"])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(
            [np.datetime64(t) for t in time_strings_interp_to]
        )
        input_times_float64 = input_times_npdt64.astype(np.float64)
        interp_to_time_float64 = interp_to_times_npdt64.astype(np.float64)
        result = np.interp(interp_to_time_float64, input_times_float64, values_input)
        # This works, unlike scipy and xarray!
        self.assertTrue((result >= 0.0).all())

    def test_tinterpol_nonnegative2(self):
        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)
        time_strings_interp_to = np.array(["2018-07-01T13:02:16.922475008"])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(
            [np.datetime64(t) for t in time_strings_interp_to]
        )
        store_data("interp_input", data={"x": input_times_npdt64, "y": values_input})
        store_data("interp_to", data={"x": interp_to_times_npdt64, "y": [0.0]})
        tinterpol("interp_input", "interp_to", newname="interp_result")
        result = get_data("interp_result")
        # This is known to fail, apparently due to limitations of scipy.interpolate.interp1d which are
        # unlikely to ever be fixed. See tests above for xarray.interp and scipy
        # self.assertTrue((result.y >= 0.0).all())
        self.assertAlmostEqual(result.y[0], 0.0, places=6)

    def test_tinterpol_slinear(self):
        # This test uses the 'slinear' interpolation method (order 1 spline) which
        # seems not to be susceptible to the issue that method='linear' has with interpolating
        # to points exactly matching the input times.
        time_strings_input = np.array(
            [
                "2018-07-01T13:02:16.892474880",
                "2018-07-01T13:02:16.922475008",
                "2018-07-01T13:02:16.952474880",
            ]
        )
        values_input = np.array([0.028584518, 0.0, 0.013626526], dtype=np.float32)
        time_strings_interp_to = np.array(["2018-07-01T13:02:16.922475008"])
        input_times_npdt64 = np.array([np.datetime64(t) for t in time_strings_input])
        interp_to_times_npdt64 = np.array(
            [np.datetime64(t) for t in time_strings_interp_to]
        )
        store_data("interp_input", data={"x": input_times_npdt64, "y": values_input})
        store_data("interp_to", data={"x": interp_to_times_npdt64, "y": [0.0]})
        tinterpol(
            "interp_input", "interp_to", newname="interp_result", method="slinear"
        )
        result = get_data("interp_result")
        self.assertTrue((result.y >= 0.0).all())


if __name__ == "__main__":
    unittest.main()



"""Transform SSL data to DSL data.

Notes:
    Works in a similar way to IDL spedas ssl2gse.pro
"""

import logging
from math import pi
import numpy as np
from copy import deepcopy

from pyspedas.tplot_tools import get_data, store_data, data_exists, get_coords, set_coords
from pyspedas.projects.themis.state_tools import Spinmodel,get_spinmodel
from pyspedas.projects.themis import autoload_support


def ssl2dsl(name_in: str, name_out: str, isdsltossl: bool = False, ignore_input_coord: bool  = False,
            probe: str=None, use_spinphase_correction: bool=True, eclipse_correction_level: int=0) -> int:
    """Transform ssl to dsl.

    Parameters
    ----------
        name_in: str
            Name of input tplot variable (e.g. 'tha_fgl_ssl')
        name_out: str
            Name of output tplot variable (e.g. 'tha_fgl_dsl')
        isdsltossl: bool
            If 0 (default) then SSL to DSL.
            If 1, then DSL to SSL.
        ignore_input_coord: bool
            if False (default), then fail and return 0 if input coordinate system does not match requested transform
            if True, do not check input coordinate system.
        probe: str
            Usually optional, if the variable name is prefixed with 'tha', 'thb', etc.
            Otherwise, one of ['a','b','c','d','e','f']
        use_spinphase_correction: bool
            If True (default), use spin phase corrections from V03 STATE CDF
            if False, omit this
        eclipse_correction_level: int
            Specify which of the three available spin models to use for this transform
            0: (default) No eclipse correction
            1: Eclipse corrections for waveform data
            2: Eclipse corrections for particles and spin fits (includes additional angular offset)

    Returns
    -------
        1 for successful completion.

    """
    needed_vars = [name_in]
    c = [value for value in needed_vars if data_exists(value)]
    if len(c) < 1:
        logging.error("Variables needed: " + str(needed_vars))
        m = [value for value in needed_vars if value not in c]
        logging.error("Variables missing: " + str(m))
        logging.error("Please load missing variables.")
        return 0

    if probe is None:
        probe=name_in[2]

    autoload_support(varname=name_in, probe=probe, spinmodel=True)
    spinmodel_obj=get_spinmodel(probe=probe, correction_level=eclipse_correction_level)

    if not ignore_input_coord:
        in_coord=get_coords(name_in)
        if in_coord is None:
            in_coord = "None"
        if isdsltossl and (in_coord.upper() != 'DSL'):
            logging.error("DSL to SSL transform requested, but input coordinate system is " + in_coord)
            return 0
        if not isdsltossl and (in_coord.upper() != 'SSL'):
            logging.error("SSL to DSL transform requested, but input coordinate system is " + in_coord)
            return 0

    # Get data
    result = get_data(name_in)
    in_times = result.times
    data_in = result.y
    metadata = get_data(name_in, metadata=1)
    meta_copy = deepcopy(metadata)

    logging.info('Using spin model to calculate phase versus time...')
    result = spinmodel_obj.interp_t(in_times, use_spinphase_correction=use_spinphase_correction)
    spinmodel_phase = result.spinphase * pi / 180.0
    phase = spinmodel_phase
    d0 = data_in[:, 0]
    d1 = data_in[:, 1]
    d2 = data_in[:, 2]
    out_d2 = d2

    # if isdsltossl == 0:
    #     # despin
    #     out_d0 = d0 * np.cos(phase) - d1 * np.sin(phase)
    #     out_d1 = d0 * np.sin(phase) + d1 * np.cos(phase)
    # else:
    #     # spin
    #     out_d0 = d0 * np.cos(phase) + d1 * np.sin(phase)
    #     out_d1 = -d0 * np.sin(phase) + d1 * np.cos(phase)

    out_coord = 'DSL'
    if isdsltossl:
        # despin
        phase = -1.0*phase
        out_coord = 'SSL'

    out_d0 = d0 * np.cos(phase) - d1 * np.sin(phase)
    out_d1 = d0 * np.sin(phase) + d1 * np.cos(phase)

    dd_out = [out_d0, out_d1, out_d2]
    data_out = np.column_stack(dd_out)
    store_data(name_out, data={'x': in_times, 'y': data_out}, attr_dict=meta_copy)
    set_coords(name_out,out_coord)

    return 1



from pyspedas.projects.themis.load import load


def ssc(
    trange=["2017-03-23", "2017-03-24"],
    probe="c",
    level="l2",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=True,
):
    """
    Load THEMIS current/past orbit data from CDAWeb/SSCWeb (Satellite Situation Center).

    For example: https://cdaweb.gsfc.nasa.gov/pub/data/themis/thc/ssc/

    See also, pyspedas.projects.themis.state_tools.ssc_pre

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-23', '2017-03-24']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')
            Default: 'c'

        level: str
            Data type; Default: 'l2'; Unused.

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False; only loads data with a "VAR_TYPE" attribute of "data"

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None; all variables are loaded

        varnames: list of str
            List of variable names to load
            Default: Empty list, so all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword
            Default: True

    Returns
    -------
    List of str
        List of tplot variables created
        Empty list if no data

    Example
    -------
        >>> from pyspedas.projects.themis import ssc
        >>> vars = ssc(probe='d', trange=['2013-11-5', '2013-11-6'])
        >>> print(vars)


    """
    return load(
        instrument="ssc",
        trange=trange,
        level=level,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        probe=probe,
        time_clip=time_clip,
        no_update=no_update,
    )



import pyspedas
from pyspedas.tplot_tools import store_data, get_data, tnames
import numpy as np
import copy
import logging


def degap(
    tvar,
    dt=None,
    margin=0.25,
    maxgap=None,
    func="nan",
    newname=None,
    onenanpergap=False,
    twonanpergap=False,
):
    """
    Fills gaps in the data either with NaNs or the last number.

    Parameters
    ----------
        tvar : str or list[str]
            Names of tplot variables to degap (wildcards accepted)
        dt : int/float
            Step size of the data in seconds, default is to use the median time interval
        margin : int/float, optional, default is 0.25 seconds
            The maximum deviation from the step size allowed before degapping occurs.  In other words, if you'd like to fill in data every 4 seconds
            but occasionally the data is 4.1 seconds apart, set the margin to .1 so that a data point is not inserted there.
        maxgap : int|float, optional
            Maximum gap length (in seconds) that will be filled.  If None, defaults to entire time range (i.e.
            all gaps with length > (dt+margin) will be filled)
        func : str, optional
            Either 'nan' or 'ffill', which overrides normal interpolation with NaN
            substitution or forward-filled values.
        newname : str, optional
            The new tplot variable name to store the data into.  If None, then the data is overwritten.
            This is not an option for multiple variable input, for multiple or pseudo variables, the data is overwritten.
        onenanpergap : bool
            if set to True, then only insert one NaN value, rather than adding NaN values at dt resolution
        twonanpergap : bool
            if set to True, then only insert one NaN value, rather than adding NaN values at dt resolution
    Returns
    -------
        None
            Creates a new tplot variable with the degap data

    Examples
    --------

        >>> import pyspedas
        >>> time = [pyspedas.time_float("2020-01-01") + i for i in [1, 2, 3, 4, 5, 6, 9, 10, 11]]
        >>> y = [1, 2, 3, 4, 5, 6, 9, 10, 11]
        >>> pyspedas.store_data("a", data={"x": time, "y": y})
        >>> degap("a", newname="b")
        >>> b = pyspedas.get("b")
        >>> print(b)
    """

    # check for globbed or array input, and call recursively
    tn = tnames(tvar)
    if len(tn) == 0:
        return
    elif len(tn) > 1:
        for j in range(len(tn)):
            degap(
                tn[j],
                dt=dt,
                margin=margin,
                func=func,
                onenanpergap=onenanpergap,
                twonanpergap=twonanpergap,
            )
        return

    # here we have 1 variable

    # fix from T.Hori, 2023-04-10, jimm02
    #    gap_size = np.diff(pyspedas.tplot_tools.data_quants[tvar].coords['time']) This is in Nanoseconds, and causes a type mismatch with dt+margin
    #    new_tvar_index = pyspedas.tplot_tools.data_quants[tvar].coords['time']
    new_tvar_index = get_data(tvar)[0]  # Unix time float64
    gap_size = np.diff(new_tvar_index)
    if maxgap is None:
        maxgap = np.nanmax(new_tvar_index)-np.nanmin(new_tvar_index)

    # Default for dt is the median value of gap_size, the time interval differences
    if dt == None:
        dt = np.median(gap_size)

    gap_index_locations = np.where((gap_size > dt + margin) & (gap_size < maxgap))
    values_to_add = np.array([])
    if onenanpergap == True:
        for i in gap_index_locations[0]:
            values_to_add = np.append(values_to_add, new_tvar_index[i] + dt)
    elif twonanpergap == True:
        # add two NaN values between the two values, either at margin if it's nonzero, or at dt/2
        # since the gap is greater than dt, this will work
        if margin > 0.0:
            if margin < dt / 2.0:
                dt_nan = margin
            else:
                dt_nan = dt / 2.0
        else:
            dt_nan = dt / 2.0
        for i in gap_index_locations[0]:
            values_to_add = np.append(values_to_add, new_tvar_index[i] + dt_nan)
            values_to_add = np.append(values_to_add, new_tvar_index[i + 1] - dt_nan)
    else:
        for i in gap_index_locations[0]:
            values_to_add = np.append(
                values_to_add, np.arange(new_tvar_index[i], new_tvar_index[i + 1], dt)
            )

    # new_index = np.sort(np.unique(np.concatenate((values_to_add, new_tvar_index))))
    new_index_float64 = np.sort(
        np.unique(np.concatenate((values_to_add, new_tvar_index)))
    )

    if func == "nan":
        method = None
    if func == "ffill":
        method = "ffill"

    # Replace any NaN or inf time values with 0.0  (Is this needed? It seems like it could result in non-monotonic times)
    cond=np.logical_not(np.isfinite(new_index_float64))
    new_index_float64[cond]=0.0

    # Convert back to datetime64 (nanoseconds since epoch)
    new_index=np.array(new_index_float64*1e9,dtype='datetime64[ns]')

    # This can fail if the stored quantities are np.datetime64 and new_index is something else, like datetime.datetime
    a = pyspedas.tplot_tools.data_quants[tvar].reindex({"time": new_index}, method=method)

    if newname is None:
        a.name = tvar
        a.attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
        pyspedas.tplot_tools.data_quants[tvar] = copy.deepcopy(a)
    else:
        if "spec_bins" in a.coords:
            store_data(
                newname,
                data={"x": a.coords["time"], "y": a.values, "v": a.coords["spec_bins"]},
            )
            pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(
                pyspedas.tplot_tools.data_quants[tvar].attrs
            )
        else:
            store_data(newname, data={"x": a.coords["time"], "y": a.values})
            pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(
                pyspedas.tplot_tools.data_quants[tvar].attrs
            )

    return



import unittest

import numpy as np

import pyspedas
from pyspedas import (
    data_exists,
    time_double,
    tinterpol,
    join_vec,
    store_data,
    get_data,
    tdeflag,
    del_data,
    set_coords,
)
from pyspedas.geopack import tt89, tt96, tt01, tts04, trace_equator_89, trace_iono_89
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas.geopack.get_w_params import get_w

trange = ["2015-10-16", "2015-10-17"]


def gen_circle():
    # Generate a circle at 5 RE in the XZ plane
    angle = np.arange(0.0, 361.0, 1.0)
    angle_rad = np.deg2rad(angle)
    y = np.zeros(len(angle_rad), np.float64)
    x = 5.0 * np.sin(angle_rad) * 6371.2
    z = 5.0 * np.cos(angle_rad) * 6371.2
    t = np.zeros(len(angle_rad))
    t[:] = time_double("2024-01-01/06:31:00") + np.arange(0.0, 361.0, 1.0)
    pos = np.zeros((len(angle_rad), 3), np.float64)
    pos[:, 0] = x
    pos[:, 1] = y
    pos[:, 2] = z
    store_data("circle_magpoles_5re", data={"x": t, "y": pos})
    set_coords("circle_magpoles_5re", "GSM")


def get_params(model, g_variables=None):
    support_trange = [
        time_double(trange[0]) - 60 * 60 * 24,
        time_double(trange[1]) + 60 * 60 * 24,
    ]
    pyspedas.projects.kyoto.dst(trange=support_trange)
    pyspedas.projects.omni.data(trange=trange)
    join_vec(["BX_GSE", "BY_GSM", "BZ_GSM"])
    if model == "t01" and g_variables is None:
        g_variables = [6.0, 10.0]
    else:
        if g_variables is not None:
            if not isinstance(g_variables, str) and not isinstance(
                g_variables, np.ndarray
            ):
                g_variables = None
    return get_tsy_params(
        "kyoto_dst",
        "BX_GSE-BY_GSM-BZ_GSM_joined",
        "proton_density",
        "flow_speed",
        model,
        pressure_tvar="Pressure",
        g_variables=g_variables,
        speed=True,
    )


class LoadTestCases(unittest.TestCase):
    def test_igrf(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        tt89("mms1_mec_r_gsm", igrf_only=True)
        self.assertTrue(mec_vars)
        self.assertTrue(data_exists("mms1_mec_r_gsm_bt89"))

    def test_tt89(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        tt89("mms1_mec_r_gsm")
        self.assertTrue(mec_vars)
        self.assertTrue(data_exists("mms1_mec_r_gsm_bt89"))

    def test_tt96(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        self.assertTrue(mec_vars)
        params = get_params("t96")
        # This interpolation can result in NaNs in the position variable, so they need to be cleaned
        tinterpol("mms1_mec_r_gsm", "proton_density")
        tdeflag("mms1_mec_r_gsm-itrp", newname="mms1_clean")
        tt96("mms1_clean", parmod=params)
        self.assertTrue(data_exists("mms1_clean_bt96"))

    def test_tt01(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        self.assertTrue(mec_vars)
        params = get_params("t01")
        # This can yield nans in the interpolated position variable for times outside the range of proton_density.
        # We don't want to pass NaNs to any of the geopack routines, so deflag
        tinterpol("mms1_mec_r_gsm", "proton_density")
        tdeflag("mms1_mec_r_gsm-itrp", newname="mms1_clean")
        # No gvariables passed, defaults to two constants which will be replicated to the correct number of points
        tt01("mms1_clean", parmod=params)
        self.assertTrue(data_exists("mms1_clean_bt01"))
        del_data("mms1_clean_bt01")
        mec = get_data("mms1_clean")
        gvars = np.zeros((len(mec.times), 2))
        # Note that mec.times may now have a different number of points than proton_density!
        gvars[:, 0] = np.repeat(6.0, len(mec.times))
        gvars[:, 1] = np.repeat(10.0, len(mec.times))
        store_data("g_variables", data={"x": mec.times, "y": gvars})
        params = get_params("t01", g_variables="g_variables")
        # Passing g_variables as tplot variable. It will get interpolated onto the correct number of points.
        tt01("mms1_clean", parmod=params)
        self.assertTrue(data_exists("mms1_clean_bt01"))
        del_data("mms1_clean_bt01")
        # Now we'll pass g_variables as an array. No interpolation will be done, so its size
        # needs to match proton_density, not mms1_clean!
        pd = get_data("proton_density")
        gvars = np.zeros((len(pd.times), 2))
        gvars[:, 0] = np.repeat(6.0, len(pd.times))
        gvars[:, 1] = np.repeat(10.0, len(pd.times))
        params = get_params("t01", g_variables=gvars)
        tt01("mms1_clean", parmod=params)
        self.assertTrue(data_exists("mms1_clean_bt01"))

    def test_tts04(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        self.assertTrue(mec_vars)
        params = get_params("ts04")
        tinterpol("mms1_mec_r_gsm", "proton_density")
        tts04("mms1_mec_r_gsm-itrp", parmod=params)
        self.assertTrue(data_exists("mms1_mec_r_gsm-itrp_bts04"))

    def test_get_w(self):
        w_vals = get_w(trange=["2015-10-16", "2015-10-17"])
        self.assertTrue(w_vals)

    def test_errors(self):
        # exercise some of the error code
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        self.assertTrue(mec_vars)
        params = get_params("ts04")
        self.assertTrue(params)
        tinterpol("mms1_mec_r_gsm", "proton_density")
        tts04("var_doesnt_exist")
        tts04("mms1_mec_r_gsm-itrp", parmod=None)
        tt01("var_doesnt_exist")
        tt01("mms1_mec_r_gsm-itrp", parmod=None)
        tt96("var_doesnt_exist")
        tt96("mms1_mec_r_gsm-itrp", parmod=None)
        tt89("var_doesnt_exist")
        invalidmodel = get_params("89")
        self.assertIsNone(invalidmodel)
        invalidg = get_params("t01", g_variables=1)
        self.assertIsNone(invalidg)
        invalidg2 = get_params("t01", g_variables="g_vars")
        self.assertIsNone(invalidg2)
        notrange = get_w()  # no trange
        self.assertIsNone(notrange)
        invalidtrange = get_w(trange=["2050-01-01", "2050-01-02"])
        self.assertIsNone(invalidtrange)

    def test_t89_equ_n(self):
        trace_equator_89(
            time_double("2007-03-23/00:00:00"), np.array([-2.0, 0.0, 1.0]), iopt=3
        )

    def test_t89_equ_s(self):
        trace_equator_89(
            time_double("2007-03-23/00:00:00"), np.array([-2.0, 0.0, -1.0]), iopt=3
        )

    def test_t89_iono_n_n(self):
        trace_iono_89(
            time_double("2007-03-23/00:00:00"), np.array([-2.0, 0.0, 1.0]), iopt=3
        )

    def test_t89_iono_n_s(self):
        trace_iono_89(
            time_double("2007-03-23/00:00:00"),
            np.array([-2.0, 0.0, 1.0]),
            iopt=3,
            south=True,
        )

    def test_t89_iono_s_n(self):
        trace_iono_89(
            time_double("2007-03-23/00:00:00"), np.array([-2.0, 0.0, -1.0]), iopt=3
        )

    def test_t89_iono_s_s(self):
        trace_iono_89(
            time_double("2007-03-23/00:00:00"),
            np.array([-2.0, 0.0, -1.0]),
            iopt=3,
            south=True,
        )

    def test_t96_roi(self):
        gen_circle()
        params = get_params("t96")
        dat = get_data(params)
        circ_dat = get_data("circle_magpoles_5re")
        n = len(circ_dat.times)
        newdat = np.zeros((n, 10), np.float64)
        newdat[:, :] = dat.y[0,]
        store_data(params, data={"x": circ_dat.times, "y": newdat})
        tt96("circle_magpoles_5re", parmod=params)
        # pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bt96'])

    def test_t01_roi(self):
        gen_circle()
        params = get_params("t01")
        dat = get_data(params)
        circ_dat = get_data("circle_magpoles_5re")
        n = len(circ_dat.times)
        newdat = np.zeros((n, 10), np.float64)
        newdat[:, :] = dat.y[0,]
        store_data(params, data={"x": circ_dat.times, "y": newdat})
        tt01("circle_magpoles_5re", parmod=params)
        # pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bt01'])

    def test_t04_roi(self):
        gen_circle()
        params = get_params("ts04")
        dat = get_data(params)
        circ_dat = get_data("circle_magpoles_5re")
        n = len(circ_dat.times)
        newdat = np.zeros((n, 10), np.float64)
        newdat[:, :] = dat.y[0,]
        store_data(params, data={"x": circ_dat.times, "y": newdat})
        tts04("circle_magpoles_5re", parmod=params)
        # pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bts04'])


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timezone
import numpy as np
from pyspedas import time_string, time_datetime, time_double, degap, store_data, get_data, options, data_exists, time_ephemeris
from numpy.testing import assert_allclose

class TimeTestCases(unittest.TestCase):
    def test_time_datetime(self):
        """Test time_datetime function."""
        now = time_datetime()
        self.assertTrue(isinstance(now, datetime))
        self.assertTrue(time_datetime("2015-12-15/00:00") == datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc))
        self.assertTrue(time_datetime(1450137600.0000000) == datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc))
        self.assertTrue(
            [time_datetime(1450137600.0000000), time_datetime(1444953600.0000000)]
            == [datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc), datetime(2015, 10, 16, 0, 0, tzinfo=timezone.utc)]
        )
        self.assertTrue(
            time_datetime([1450137600.0000000, 1444953600.0000000])
            == [datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc), datetime(2015, 10, 16, 0, 0, tzinfo=timezone.utc)]
        )

    def test_numpy_time_get_store(self):
        trange = time_double(np.array(['2007-03-23', '2007-03-24']))
        trange_np = np.array(trange)
        # Create the same timestamps in three different time resolutions
        np_sec = np.array(trange_np,dtype="datetime64[s]")
        np_us = np_sec.astype('datetime64[us]')
        np_ns = np_sec.astype("datetime64[ns]")
        print('original trange:', trange)
        print('trange as numpy array: ',trange_np)
        print('npdatetime64[s]: ', np_sec)
        print('npdatetime64[us]: ', np_us)
        print('npdatetime64[ns]: ', np_ns)
        data = [1.0, 2.0]
        store_data('tvar_sec', data={'x': np_sec, 'y': data})
        store_data('tvar_us',  data={'x': np_us,  'y': data})
        store_data('tvar_ns',  data={'x': np_ns,  'y': data})
        # Check that different resolutions return the same unix times
        tv_secdata = get_data('tvar_sec')
        tv_usdata = get_data('tvar_us')
        tv_nsdata = get_data('tvar_ns')
        print('max time converted from sec: ', np.max(tv_secdata.times))
        print('max time converted from us: ', np.max(tv_usdata.times))
        print('max time converted from ns: ', np.max(tv_nsdata.times))
        assert_allclose(trange_np, tv_nsdata.times)
        assert_allclose(trange_np, tv_secdata.times)
        assert_allclose(trange_np, tv_usdata.times)
        # plot_options["trange"] is derived from the input times, need to check that too
        tv_secmd = get_data('tvar_sec',metadata=True)
        tv_usmd = get_data('tvar_us',metadata=True)
        tv_nsmd = get_data('tvar_ns',metadata=True)
        assert_allclose(tv_secmd['plot_options']['trange'],trange_np)
        assert_allclose(tv_usmd['plot_options']['trange'],trange_np)
        assert_allclose(tv_nsmd['plot_options']['trange'],trange_np)


    def test_time_string(self):
        """Test time_string function."""
        self.assertTrue(time_string(fmt="%Y-%m-%d") == datetime.now().strftime("%Y-%m-%d"))
        self.assertTrue(time_string(1450181243.767, fmt="%Y-%m-%d") == "2015-12-15")
        self.assertTrue(time_string(1450181243.767, fmt="%Y-%m-%d/%H:%M:%S") == "2015-12-15/12:07:23")
        self.assertTrue(time_string(1450181243.767) == "2015-12-15 12:07:23.767000")
        self.assertTrue(
            time_string([1450181243.767, 1450181263.767]) == ["2015-12-15 12:07:23.767000", "2015-12-15 12:07:43.767000"]
        )

    def test_time_double(self):
        """Test time_double function."""
        self.assertTrue(time_string(time_double(), fmt="%Y-%m-%d") == datetime.now().strftime("%Y-%m-%d"))
        self.assertTrue(time_double("2015-12-15/12:00") == 1450180800.0000000)
        self.assertTrue(time_double("2015-12-15/12") == 1450180800.0000000)
        # self.assertTrue(time_double('2015-12-15/6') == 1450159200.0000000) #this one doesn't work
        self.assertTrue(time_double("2015-12-15/6:00") == 1450159200.0000000)
        self.assertTrue(time_double("2015-12-15/06:00") == 1450159200.0000000)
        self.assertTrue(time_double("2015-12-15") == 1450137600.0000000)
        self.assertTrue(time_double("2015 12 15") == 1450137600.0000000)
        self.assertTrue(time_double("2015-12") == 1448928000.0000000)
        self.assertTrue(time_double("2015") == 1420070400.0000000)
        self.assertTrue(time_double("2015-12-15 12:07:23.767000") == 1450181243.767)
        self.assertTrue(
            time_double(["2015-12-15 12:07:23.767000", "2015-12-15 12:07:43.767000"]) == [1450181243.767, 1450181263.767]
        )

    def test_degap(self):
        float_times = np.array([1.0, 2.0, 3.0, 11.0, 12.0, 13.0, 21.0, 22.0, 23.0])
        int_times = np.array(float_times, dtype=np.int64)
        str_times = time_string(float_times)
        dtimes = np.array(float_times * 1e9, dtype="datetime64[ns]")
        data = [1.0, 2.0, 3.0, 11.0, 12.0, 13.0, 21.0, 22.0, 23.0]
        sp = np.array([1, 2, 3])
        spdata = [sp, sp, sp, sp, sp, sp, sp, sp, sp]
        spv = [100, 200, 300]
        # Scalar data with floating point times
        store_data("degap_test_float", data={"x": float_times, "y": data})
        # Spectral data with floating point times
        store_data("degap_test_float_spec", data={"x": float_times, "y": spdata, "v": spv})
        options("degap_test_float_spec", "spec", 1)
        # Scalar data with integer times
        store_data("degap_test_int", data={"x": int_times, "y": data})
        # Spectral data with integer times
        store_data("degap_test_int_spec", data={"x": int_times, "y": spdata, "v": spv})
        options("degap_test_int_spec", "spec", 1)
        # Scalar data with datetime64[ns] times
        store_data("degap_test_dt64", data={"x": dtimes, "y": data})
        # Spectral data with datetime64[ns] times
        store_data("degap_test_dt64_spec", data={"x": dtimes, "y": spdata, "v": spv})
        options("degap_test_dt64_spec", "spec", 1)
        # Scalar data with string times
        store_data("degap_test_string", data={"x": str_times, "y": data})
        # Spectral data with string times
        store_data("degap_test_string_spec", data={"x": str_times, "y": spdata, "v": spv})
        options("degap_test_string_spec", "spec", 1)

        degap("degap_test_float", dt=2.0, maxgap=5.0, newname="degap_maxgap")
        dg_data = get_data("degap_maxgap")
        # Should NOT insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 11.0)
        self.assertTrue(np.isfinite(dg_data.y[3]))
        self.assertTrue(len(dg_data.times), 9)

        degap("degap_test_float", dt=2.0, newname="degap_default_gapfill")
        dg_data = get_data("degap_default_gapfill")
        # Should NOT insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertTrue(len(dg_data.times), 15)

        degap("degap_test_float", dt=2.0, onenanpergap=True, newname="degap_onenanpergap")
        dg_data = get_data("degap_onenanpergap")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertEqual(dg_data.times[7], 15.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertFalse(np.isfinite(dg_data.y[7]))
        self.assertTrue(len(dg_data.times), 11)

        degap("degap_test_float", dt=2.0, twonanpergap=True, newname="degap_twonanpergap")
        dg_data = get_data("degap_twonanpergap")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 3.25)
        self.assertEqual(dg_data.times[4], 10.75)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertFalse(np.isfinite(dg_data.y[4]))
        self.assertTrue(len(dg_data.times), 13)

        degap("degap_test_float", dt=2.0)
        dg_data = get_data("degap_test_float")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap("degap_test_int", dt=2.0)
        dg_data = get_data("degap_test_int")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap("degap_test_float_spec", dt=2.0)
        dg_data = get_data("degap_test_float_spec")
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap("degap_test_int_spec", dt=2.0)
        dg_data = get_data("degap_test_int_spec")
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap("degap_test_dt64", dt=2.0)
        dg_data = get_data("degap_test_dt64")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap("degap_test_dt64_spec", dt=2.0)
        dg_data = get_data("degap_test_dt64_spec")
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap("degap_test_string", dt=2.0)
        dg_data = get_data("degap_test_string")
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap("degap_test_string_spec", dt=2.0)
        dg_data = get_data("degap_test_string_spec")
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

    def test_missing_pseudovar(self):
        times = np.array([1.0, 2.0, 3.0, 11.0, 12.0, 13.0, 21.0, 22.0, 23.0])
        data = [1.0, 2.0, 3.0, 11.0, 12.0, 13.0, 21.0, 22.0, 23.0]

        store_data("basevar1", data={"x": times, "y": data})
        store_data("basevar2", data={"x": times, "y": data})
        store_data("pseudovar_missing", data=["basevar1", "basevar2", "missing_var"])
        self.assertTrue(data_exists("pseudovar_missing"))
        store_data("pseudovar_empty_list", data=[])
        self.assertFalse(data_exists("pseudovar_empty_list"))
        store_data("pseudovar_all_bad", data=["missing_var1", "missing_var2", "missing_var3"])
        self.assertFalse(data_exists("pseudovar_all_bad"))

    def test_time_ephemeris(self):
        testdate = "2025-01-01/00:00:00"
        ut = time_double(testdate)
        et = time_ephemeris(ut)
        self.assertAlmostEqual(ut, 1735689600.0000000, places=5)
        self.assertAlmostEqual(et, 788961669.18399811, places=5)
        et_to_ut = time_ephemeris(et, et2ut=True)
        self.assertAlmostEqual(ut, et_to_ut, places=6)


if __name__ == "__main__":
    unittest.main()



import pyspedas
from pyspedas.tplot_tools import tplot_wildcard_expand
import logging

def highlight(variables=None, range=None, color='gray', alpha=0.2, fill=True, edgecolor=None, facecolor=None, hatch=None, delete=False):
    """
    Highlight a time interval on tplot variables by adding shading to the plot panel

    Most parameters are passed to the matplotlib axes.axvspan method via a dictionary.

    Parameters
    ==========

    variables: str or list
        tplot variables to add highlights to (Wildcards accepted)
    range: array of floats
        Start and end of highlight time interval, as Unix times
    color: str
        Color to use for the highlight
    alpha: float
        Transparency of highlight
    fill: bool
        Fill color of highlight
    edgecolor: str
        Edge color of highlight
    facecolor: str
        Face color of highlight
    hatch: str
        Hatch pattern to use for highlight
    delete: bool
        If True, delete all highlights associated with the specified variables

    """
    if not isinstance(variables, list):
        variables = [variables]
    tvars=tplot_wildcard_expand(variables)
    if len(tvars) == 0:
        logging.warning("highlight: No valid tplot names specified")

    for variable in tvars:
        if delete:
            pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['highlight_intervals'] = None
            continue
        if range is None:
            pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['highlight_intervals'] = None
            continue
        interval = {'location': range,
                    'color': color,
                    'alpha': alpha,
                    'fill': fill,
                    'edgecolor': edgecolor,
                    'facecolor': facecolor,
                    'hatch': hatch}
        if pyspedas.tplot_tools.data_quants[variable].attrs['plot_options'].get('highlight_intervals') is None:
            pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['highlight_intervals'] = [interval]
        else:
            pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['highlight_intervals'].append(interval)




import math
import numpy as np
from scipy.ndimage import shift

from copy import deepcopy

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum

def erg_pgs_make_phi_spec(data_in, resolution=None, no_ang_weighting=False):
    """
    Builds phi (longitudinal) spectrogram from the particle data structure

    Input:
        data_in: dict
            Particle data structure

    Parameters:
        resolution: int
            Number of phi bins in the output

    Returns:
        Tuple containing: (phi values for y-axis, spectrogram values)
    """

    dr = math.pi/180.

    data = deepcopy(data_in)

    # zero inactive bins to ensure areas with no data are represented as NaN
    data['data'] = np.where(data['bins'] == 0,0.,data['data'])

    # get number of phi values
    if resolution is None:
        # method taken from the IDL code
        idx = np.nanargmin(np.abs((data['theta'][0, :])))
        n_phi = len(np.argwhere(data['theta'][0, :] == np.abs((data['theta'][0, :]))[idx]))
    else:
        n_phi = resolution

    # init this sample's piece of the spectrogram
    ave = np.zeros(n_phi)

    # form grid specifying the spectrogram's phi bins
    phi_grid = np.linspace(0, 360.0, n_phi+1)
    phi_grid_width = np.nanmedian(phi_grid - shift(phi_grid, 1))

    # get min/max of all data bins
    # keep phi in [0, 360]
    phi_min = (data['phi'] - 0.5*data['dphi']) 
    phi_max = (data['phi'] + 0.5*data['dphi']) % 360.0
    phi_array = deepcopy(data['phi'])

    # algorithm below assumes maximums at 360 not wrapped to 0
    phi_max[phi_max == 0] = 360.0

    # keep phi in [0, 360]
    phi_min[phi_min < 0] = phi_min[phi_min < 0] + 360

    # keep track of bins that span phi=0
    wrapped = phi_min > phi_max

    # When averaging data bins will be weighted by the solid angle of their overlap,
    # with the given spectrogram bin.  Since each spectrogram bin spans all theta 
    # values the theta portion of that calculation can be done in advance.  These
    # values will later be multiplied by the overlap along phi to get the total 
    # solid angle. 
    omega_part = np.abs(np.sin(dr * (data['theta'] + .5*data['dtheta'])) - np.sin(dr * (data['theta'] - .5*data['dtheta'])))

    omega_part_flat = omega_part.flatten(order='F')
    phi_max_flat = phi_max.flatten(order='F')
    phi_min_flat = phi_min.flatten(order='F')
    wrapped_flat = wrapped.flatten(order='F')
    data_flat = data['data'].flatten(order='F')
    bins_flat = data['bins'].flatten(order='F')
    dphi_flat = data['dphi'].flatten(order='F')

    # Loop over each phi bin in the spectrogram and determine which data bins
    # overlap.  All overlapping bins will be weighted according to the solid 
    # angle of their intersection and averaged.
    for i in range(0, n_phi):

        if not no_ang_weighting:
            weight = np.zeros(phi_min_flat.shape)

            # data bins whose maximum overlaps the current spectrogram bin
            idx_max = np.argwhere((phi_max_flat > phi_grid[i]) & (phi_max_flat < phi_grid[i+1]))
            if idx_max.size != 0:
                weight[idx_max] = (phi_max_flat[idx_max] - phi_grid[i]) * omega_part_flat[idx_max]

            # data bins whose minimum overlaps the current spectrogram bin
            idx_min = np.argwhere((phi_min_flat > phi_grid[i]) & (phi_min_flat < phi_grid[i+1]))
            if idx_min.size != 0:
                weight[idx_min] = (phi_grid[i+1] - phi_min_flat[idx_min]) * omega_part_flat[idx_min]

            # data bins contained within the current spectrogram bin
            contained = np.intersect1d(idx_max, idx_min)
            if contained.size != 0:
                weight[contained] = dphi_flat[contained] * omega_part_flat[contained]
            
            # data bins that completely cover the current spectrogram bin 
            idx_all = np.argwhere(((phi_min_flat <= phi_grid[i]) & (phi_max_flat >= phi_grid[i+1])) |
                                (wrapped_flat & ((phi_min_flat > phi_grid[i+1]) & (phi_max_flat > phi_grid[i+1]))) |
                                (wrapped_flat & ((phi_min_flat < phi_grid[i]) & (phi_max_flat < phi_grid[i]))))
            if idx_all.size != 0:
                weight[idx_all] = phi_grid_width * omega_part_flat[idx_all]

            # combine indices
            idx = np.unique(np.concatenate((idx_min, idx_max, idx_all)))

            # assign a weighted average to this bin
            if idx_max.size + idx_min.size + idx_all.size > 0:
                # normalize weighting to selected, active bins
                weight[idx] = weight[idx] * bins_flat[idx]
                weight = weight/nansum(weight)

                # average
                ave[i] = nansum(data_flat[idx]*weight[idx])

        else: # ;;wighout weighting by dphi and dtheta
            id_array = np.argwhere(np.isfinite(data['data'])\
                     & (data['bins'] == 1)\
                     & (phi_array > phi_grid[i])\
                     & (phi_array < phi_grid[i+1]))

            if len(id_array) > 0:
                ave[i] = np.nanmean(data['data'][id_array[:, 0].tolist(), id_array[:, 1].tolist()])


    # get y axis
    y = (phi_grid+shift(phi_grid, 1))/2.0
    y = y[1:]

    return (y, ave)



from .gmag_isee_induction import gmag_isee_induction


from typing import List, Optional, Union

def gmag_stel_induction(
    trange: List[str] = ["2018-10-18/00:00:00", "2018-10-18/02:00:00"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    frequency_dependent: bool = False,
    force_download: bool = False,
) -> List[str]:
    """
    Load data from STEL Induction Magnetometers (wrapper for gmag_isee_induction)

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-10-18/00:00:00', '2018-10-18/02:00:00']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load. Valid values: "ath", "gak", "hus", "kap", "lcl", "mgd", "msr",
            "nai", "ptk", "rik", "sta", "zgn", "all"
            Default: ['all']

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    frequency_dependent: bool
            If set, load frequency dependent data.
            Default: False

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> ind_vars=pyspedas.projects.erg.gmag_stel_induction(trange=['2020-08-01','2020-08-02'], site='all')
    >>> tplot('isee_induction_db_dt_msr')

    """

    return gmag_isee_induction(
        trange=trange,
        suffix=suffix,
        site=site,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        no_update=no_update,
        uname=uname,
        passwd=passwd,
        time_clip=time_clip,
        ror=ror,
        frequency_dependent=frequency_dependent,
        force_download=force_download,
    )



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from .rfs import rfs_variables_to_load
from .config import CONFIG

def load(trange=['2018-11-5', '2018-11-6'],
         instrument='fields',
         datatype='mag_RTN',
         spec_types=None,  # for DFB AC spectral data
         level='l2',
         suffix='',
         prefix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         username=None,
         password=None,
         last_version=False,
         force_download=False,
         version=None):
    """
    This function loads Parker Solar Probe (PSP) data into tplot variables; this function is not
    meant to be called directly; instead, see the wrappers: 
        psp.fields: FIELDS data
        psp.spc: Solar Probe Cup data
        psp.spe: SWEAP/SPAN-e data
        psp.spi: SWEAP/SPAN-i data
        psp.epihi: ISoIS/EPI-Hi data
        psp.epilo: ISoIS/EPI-Lo data
        psp.epi ISoIS/EPI (merged Hi-Lo) data

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        instrument: str
            Spacecraft identifier ('fields', 'spc', 'spe', 'spi', 'epihi', 'epilo', 'epi')
            Default: 'fields'

        datatype: str
            Valid options: ['mag_RTN_4_Sa_per_Cyc',
                            'mag_RTN',
                            'mag_SC',
                            'mag_SC_1min',
                            'mag_SC_4_Sa_per_Cyc',
                            'sqtn_rfs_V1V2']
            Default: 'mag_RTN'

        spec_types: str
            Valid options: for DFB AC spectral data
            Default: None

        level: str
            Valid options: 'l2'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix. By default,
            no prefix is added.
            Default: ''

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: 'False', only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            Default: [], all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False
            
        username: str
            Username to download private data
            Default: None
        
        password: str
            Password to download private data
            Default: None
            
        last_version: bool
            Use a wildcard pattern to download the highest version available of the CDF
            Default: False
        
        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

        version: str
            If supplied, the load routine will look for this specific CDF version, rather than a wildcard pattern
            matching any version.  Valid version strings look like "vN.M" for FIELDS sqtn_rfs_v1v1 data, or 
            "vNN" for all other data types. Default: None

    Returns
    ----------
        List of tplot variables created.

    Examples
    ----------
        import pyspedas
        from pyspedas.tplot_tools import tplot
        fields_psp_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-6'])

        spc_psp_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'])

        spe_psp_vars = pyspedas.projects.psp.spe(trange=['2018-11-5', '2018-11-6'])

        spi_psp_vars = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-6'])

        epihi_psp_vars = pyspedas.projects.psp.epihi(trange=['2018-11-5', '2018-11-6'])

        epilo_psp_vars = pyspedas.projects.psp.epilo(trange=['2018-11-5', '2018-11-6'])

        epi_psp_vars = pyspedas.projects.psp.epi(trange=['2018-11-5', '2018-11-6'])

    """
    # remote path formats generally are going to be all lowercase except for
    # on the Berkeley FIELDS server
    if (username is not None) and (datatype in ['mag_RTN_1min',
                                            'mag_RTN_4_Sa_per_Cyc',
                                            'mag_RTN',
                                            'mag_SC',
                                            'mag_SC_1min',
                                            'mag_SC_4_Sa_per_Cyc',
                                            'sqtn_rfs_V1V2'
                                            ]):
        pass
    else:
        datatype = datatype.lower()

    if suffix is None:
        suffix = ''

    if prefix is not None:
        user_prefix = prefix
    else:
        user_prefix = ''

    prefix = user_prefix + 'psp_'  #To cover the case if one *does* call this routine directly.

    if version is not None:
        cdf_version = version
        last_version = False
    else:
        last_version = True
        if datatype.lower() == 'sqtn_rfs_v1v2':
            cdf_version = "v?.?"
        else:
            cdf_version = "v??"
    
    file_resolution = 24*3600.
    if instrument == 'fields':
        prefix = user_prefix + '' #CDF Variables are already prefixed with psp_fld_

        # 4_per_cycle and 1min are daily, not 6h like the full resolution 'mag_(rtn|sc)'
        if datatype in ['mag_rtn', 'mag_sc']:
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_'+cdf_version+'.cdf'
            file_resolution = 6*3600.
        elif datatype in ['mag_rtn_1min', 'mag_sc_1min', 'rfs_hfr', 'rfs_lfr', 'rfs_burst', 'f2_100bps', 'aeb']:
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype in ['mag_rtn_4_per_cycle', 'mag_rtn_4_sa_per_cyc']:
            pathformat = instrument + '/' + level + '/mag_rtn_4_per_cycle/%Y/psp_fld_' + level + '_mag_rtn_4_sa_per_cyc_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype in ['mag_sc_4_per_cycle', 'mag_sc_4_sa_per_cyc']:
            pathformat = instrument + '/' + level + '/mag_sc_4_per_cycle/%Y/psp_fld_' + level + '_mag_sc_4_sa_per_cyc_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype == 'sqtn_rfs_v1v2':
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype == 'rfs_lfr_qtn':
            pathformat = instrument + '/' + level + '/' + datatype + '/psp_fld_' + level + '_' + datatype + '_%Y%m*_'+cdf_version+'.cdf'
        elif datatype in ['dfb_dc_spec', 'dfb_ac_spec', 'dfb_dc_xspec', 'dfb_ac_xspec']:
            out_vars = []
            for item in spec_types:
                loaded_data = load(trange=trange, instrument=instrument, datatype=datatype + '_' + item, level=level, 
                    suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
                    downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, last_version=last_version, force_download=force_download)
                if loaded_data != []:
                    out_vars.extend(loaded_data)
            return out_vars
        elif datatype[:12] == 'dfb_dc_spec_' or datatype[:12] == 'dfb_ac_spec_' or datatype[:13] == 'dfb_dc_xspec_' or datatype[:13] == 'dfb_ac_xspec_':
            if datatype[:13] == 'dfb_dc_xspec_' or datatype[:13] == 'dfb_ac_xspec_':
                dtype_tmp = datatype[:12]
                stype_tmp = datatype[13:]
            else:
                dtype_tmp = datatype[:11]
                stype_tmp = datatype[12:]
            pathformat = instrument + '/' + level + '/' + dtype_tmp + '/' + stype_tmp + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype == 'sqtn_rfs_v1v2':
            # unfortunately the naming format of quasi-thermal-noise cdf file is different from others
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype == 'sqtn_rfs_V1V2':
            # unpublished QTN data
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
        elif datatype == 'merged_scam_wf':
            if username == None:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_'+cdf_version+'.cdf'
            else:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_'+cdf_version+'.cdf'

        # unpublished data
        elif username != None:
            if datatype in ['mag_RTN', 'mag_SC']:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_'+cdf_version+'.cdf'
                file_resolution = 6*3600.

            elif datatype in ['mag_RTN_1min', 'mag_RTN_4_Sa_per_Cyc', 'mag_SC_1min', 'mag_SC_4_Sa_per_Cyc']:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'

            elif datatype ==  'sqtn_rfs_V1V2':
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
            elif datatype in ['ephem_spp_rtn']:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/spp_fld_' + level + '_' + datatype + '_%Y%m%d_v01.cdf'
            else:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'


        else:
            # Generic SPDF path.  
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_'+cdf_version+'.cdf'
            file_resolution = 6*3600.

        # Files on Berkeley server are stored in monthly directories 
        if username != None:
            pathformat = pathformat.replace('/%Y/psp_fld', '/%Y/%m/psp_fld')
            if datatype == 'rfs_lfr_qtn':
                pathformat = pathformat.replace('l3/rfs_lfr_qtn/', 'l3/rfs_lfr_qtn/%Y/%m/')
            if level == 'l1':
                pathformat = pathformat.replace('psp_fld', 'spp_fld')


    elif instrument == 'spc':
        # spc changes the filename on the secure server depending on
        # whether data has been publicly released or not.
        # the publicly released data is on the NASA spdf server with the 'psp'
        # prefix. 
        #
        # However, data on the secure server exists with both the 'psp' and 'spp' 
        # prefixes. This prefix ambiguity will be handled farther down
        # and is set to 'psp' for the time being.
        prefix = user_prefix + 'psp_spc_'
        pathformat = 'sweap/spc/' + level + '/' + datatype + '/%Y/psp_swp_spc_' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'
    elif instrument == 'spe':
        prefix = user_prefix + 'psp_spe_'
        pathformat = 'sweap/spe/' + level + '/' + datatype + '/%Y/psp_swp_sp?_*_%Y%m%d_'+cdf_version+'.cdf'
    elif instrument == 'spi':
        if username is None:
            prefix = user_prefix + 'psp_spi_'
            pathformat = 'sweap/spi/' + level + '/' + datatype + '/%Y/psp_swp_spi_*_%Y%m%d_'+cdf_version+'.cdf'
        else:
            # unpublished data
            prefix = user_prefix + 'psp_spi_'
            pathformat = 'sweap/spi/' + level + '/' + datatype + '/%Y/%m/psp_swp_' + datatype + '*_%Y%m%d_v0?.cdf'
    elif instrument == 'epihi':
        prefix = user_prefix + 'psp_epihi_'
        pathformat = 'isois/epihi/' + level + '/' + datatype + '/%Y/psp_isois-epihi_' + level + '*_%Y%m%d_'+cdf_version+'.cdf'
    elif instrument == 'epilo':
        prefix = user_prefix + 'psp_epilo_'
        pathformat = 'isois/epilo/' + level + '/' + datatype + '/%Y/psp_isois-epilo_' + level + '*_%Y%m%d_'+cdf_version+'.cdf'
    elif instrument == 'epi':
        prefix = user_prefix + 'psp_isois_'
        pathformat = 'isois/merged/' + level + '/' + datatype + '/%Y/psp_isois_' + level + '-' + datatype + '_%Y%m%d_'+cdf_version+'.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange, res=file_resolution)

    out_files = []

    if username is None:
        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                        local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version, force_download=force_download)
    else:
        if instrument == 'fields':
            try:
                print("Downloading unpublished FIELDS Data....")
                files = download(
                    remote_file=remote_names, remote_path=CONFIG['fields_remote_data_dir'], 
                    local_path=CONFIG['local_data_dir'], no_download=no_update,
                    username=username, password=password, basic_auth=True,last_version=last_version, force_download=force_download,
                )
                if files == []: # I think this is a temp-fix, the logic of these blocks doesnt work now that download() doesnt raise an exception
                    raise RuntimeError("No links found.")
            except:
                files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                                local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version, force_download=force_download)
        elif instrument == 'spi':
            try:
                print("Downloading unpublished SWEAP Data....")
                files = download(
                    remote_file=remote_names, remote_path=CONFIG['sweap_remote_data_dir'], 
                    local_path=CONFIG['local_data_dir'], no_download=no_update,
                    username=username, password=password, basic_auth=True,last_version=last_version, force_download=force_download
                )
                if files == []: # I think this is a temp-fix, the logic of these blocks doesnt work now that download() doesnt raise an exception
                    raise RuntimeError("No links found.")
            except:
                files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                                local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version, force_download=force_download)

        elif instrument == 'spc':
            # try secure server first
            # secure server may have two file names, 
            # the publicly released data will be prefixed 'psp_swp_spc_' 
            # so check that filename first, as it will have the most
            # up to date version of the data

            try:
                print("Downloading unpublished SWEAP/SPC Data ('psp' prefix)....")
                prefix = user_prefix + 'psp_spc_'
                pathformat = 'sweap/spc/' + level + '/%Y/%m/psp_swp_spc_' + datatype + '_%Y%m%d_v0?.cdf'
                # find the full remote path names using the trange
                remote_names_spc = dailynames(file_format=pathformat, trange=trange, res=file_resolution)
                remote_dates = dailynames(file_format='%Y-%m-%d', trange=trange, res=file_resolution)
                files = download(
                    remote_file=remote_names_spc, remote_path=CONFIG['sweap_remote_data_dir'], 
                    local_path=CONFIG['local_data_dir'], no_download=no_update,
                    username=username, password=password, basic_auth=True,last_version=last_version, force_download=force_download
                )

                if len(files) != len(remote_dates): # I think this is a temp-fix, the logic of these blocks doesnt work now that download() doesnt raise an exception
                    if str(type(trange[1])) == "<class 'datetime.datetime'>" : #Handle case when user input is type <datetime.datetime>
                        t_end = trange[1].strftime('%Y-%m-%d')
                    trange_temp = [remote_dates[len(files)],trange[1]]    # set new trange to check for spp prefixes
                    raise RuntimeError("No 'psp...' prefix links found.")
            except RuntimeError as e:
                print(e)
                try :
                    
                    # then, if a date does not have 'psp_swp_spc_' prefix 
                    # then check for prefix 'spp_swp_spc_' which is the pre-released data
                    prefix = user_prefix + 'spp_spc_'
                    pathformat = 'sweap/spc/' + level + '/%Y/%m/spp_swp_spc_' + datatype + '_%Y%m%d_v0?.cdf'
                    # find the full remote path names using the trange
                    remote_names_spc = dailynames(file_format=pathformat, trange=trange_temp, res=file_resolution)
                    print("Downloading unpublished SWEAP/SPC Data ('spp' prefix)....")
                    files = download(
                        remote_file=remote_names_spc, remote_path=CONFIG['sweap_remote_data_dir'], 
                        local_path=CONFIG['local_data_dir'], no_download=no_update,
                        username=username, password=password, basic_auth=True,last_version=last_version, force_download=force_download
                        )
                    if files == []: # I think this is a temp-fix, the logic of these blocks doesnt work now that download() doesnt raise an exception
                        raise RuntimeError("No 'spp...' prefix links found.")
                # if all else fails switch to the NASA spdf server.        
                except RuntimeError as e :
                    print(e)
                    print("Trying public data....")
                    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                                    local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version, force_download=force_download)

    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    # find the list of varnames for RFS data
    # these files have > 1500 variables, but
    # we only load ~50
    if 'rfs' in datatype.lower() and varformat is None and varnames == []:
        varnames = rfs_variables_to_load(out_files)
        # we'll need the support data for the quality flags
        get_support_data = True

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, 
                        varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas

from pyspedas.tplot_tools import xlim


def tlimit(arg=None, full=False, last=False):
    """

    Parameters
    ----------
    arg: str or list
        If arg=='full', revert to the full time range. If arg=='last', revert to the previous time range.
        Otherwise, arg should be a list containing a start and stop time.
    full: bool
        If True, revert to the full time range. Equivalent to tlimit('full').
    last: bool:
        If True, revert to the previous time range. Equivalent to tlimit('last').

    Returns
    -------
        None

    Examples
    --------
    >>> import pyspedas
    >>> pyspedas.tlimit(['2023/03/24/', '2023/03/25'])

    """

    if full or (isinstance(arg,str) and arg == 'full'):
        if pyspedas.tplot_tools.tplot_opt_glob.get('x_range') is not None:
            del pyspedas.tplot_tools.tplot_opt_glob['x_range']
    elif last or (isinstance(arg,str) and arg == 'last'):
        pyspedas.tplot_tools.tplot_opt_glob['x_range'] = pyspedas.tplot_tools.tplot_opt_glob['x_range_last']
    else:
        minn = arg[0]
        maxx = arg[1]
        xlim(minn, maxx)
        
    return




import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data, timespan,tplot, tplot_names, get_data, options
from pyspedas.projects.erg import erg_hep_part_products
from pyspedas.projects.erg.satellite.erg.particle.erg_hep_get_dist import erg_hep_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_pgs_make_e_spec import erg_pgs_make_e_spec
from pyspedas.projects.erg.satellite.erg.particle.erg_pgs_make_theta_spec import erg_pgs_make_theta_spec
from pyspedas.projects.erg.satellite.erg.particle.erg_pgs_make_phi_spec import erg_pgs_make_phi_spec

import pyspedas
import pyspedas

display=False

class LoadTestCases(unittest.TestCase):


    def test_part_products_utilities(self):
        del_data()
        hep_vars = pyspedas.projects.erg.hep(datatype='3dflux')
        base = "erg_hep_l2_"
        for species in ['FEDU_L', 'FEDU_H']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_hep_get_dist(var, index=[0,1,2], species=None)
            theta_y, theta_v = erg_pgs_make_theta_spec(dist, resolution=dist['n_theta'],no_ang_weighting=False)
            self.assertTrue(len(theta_y) > 0)
            self.assertTrue(len(theta_v) > 0)
            self.assertTrue(len(theta_y) == len(theta_v))
            phi_y, phi_v = erg_pgs_make_phi_spec(dist, resolution=dist['n_phi'],no_ang_weighting=False)
            self.assertTrue(len(phi_y) > 0)
            self.assertTrue(len(phi_v) > 0)
            self.assertTrue(len(phi_y) == len(phi_v))

    def test_hep_theta(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', outputs='theta' )
        self.assertTrue('erg_hep_l2_FEDU_L_theta' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_theta'))
        tplot( 'erg_hep_l2_FEDU_L_theta', display=display, save_png='erg_hep_theta.png' )

    def test_hep_theta_trange(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L',trange=[ '2017-04-05 21:45:00', '2017-04-05 22:45:00'], outputs='theta' )
        self.assertTrue('erg_hep_l2_FEDU_L_theta' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_theta'))
        tplot( 'erg_hep_l2_FEDU_L_theta', display=display, save_png='erg_hep_theta_trange.png' )

    def test_hep_theta_limit_phi(self):
        del_data('*')
        # Load LEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 22:45:00'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', outputs='theta', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], phi_in=[0., 180.0] )
        tplot( 'erg_hep_l2_FEDU_L_theta', display=display, save_png='erg_hep_theta_limit_phi.png' )
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_theta'))
        self.assertTrue('erg_hep_l2_FEDU_L_theta' in vars)

    def test_hep_phi(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', outputs='phi')
        self.assertTrue('erg_hep_l2_FEDU_L_phi' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_phi'))
        tplot( 'erg_hep_l2_FEDU_L_phi', display=display, save_png='erg_hep_phi.png' )


    def test_hep_pa(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='pa' )
        self.assertTrue('erg_hep_l2_FEDU_L_pa' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_pa'))
        tplot( 'erg_hep_l2_FEDU_L_pa', display=display, save_png='erg_hep_pa.png' )

    # erg_hep_part_products doesn't support moments yet?

    def test_hep_moments(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='moments' )
        tplot(vars, display=display, save_png='erg_hep_moments.png')
        self.assertTrue('erg_hep_l2_FEDU_L_density' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_density'))
        self.assertTrue('erg_hep_l2_FEDU_L_density' in vars)
        self.assertEqual(pyspedas.get_units('erg_hep_l2_FEDU_L_density'), '1/cm^3')
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_qflux'))
        self.assertTrue('erg_hep_l2_FEDU_L_qflux' in vars)
        self.assertEqual(pyspedas.get_units('erg_hep_l2_FEDU_L_qflux'), 'eV/(cm^2-sec)')
        self.assertEqual(pyspedas.get_coords('erg_hep_l2_FEDU_L_qflux').lower(), 'dsi')


    def test_hep_fac_moments(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='fac_moments')
        tplot(vars, display=display, save_png='erg_hep_fac_moments.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_density_mag' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_density_mag'))

    def test_hep_fac_energy(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='fac_energy' )
        tplot(vars, display=display, save_png='erg_hep_fac_energy.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_energy_mag' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_energy_mag'))

    def test_hep_gyro(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='gyro' )
        tplot(vars, display=display, save_png='erg_hep_gyro.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_gyro' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_gyro'))

    def test_hep_gyro_muconv(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', mag_name=mag_vn, pos_name=pos_vn, outputs='gyro' , muconv=True)
        tplot(vars, display=display, save_png='erg_hep_gyro_muconv.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_gyro' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_gyro'))

    def test_hep_energy(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_hep_part_products( 'erg_hep_l2_FEDU_L', outputs='energy' )
        tplot( 'erg_hep_l2_FEDU_L_energy', display=display, save_png='erg_hep_en_spec.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_energy' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_energy'))

    def test_hep_energy_limit_gyro(self):
        del_data('*')
        # Load LEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux')
        vars = pyspedas.projects.erg.mgf(
            trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_hep_part_products('erg_hep_l2_FEDU_L', outputs='energy', gyro=[0., 180.],
                                     fac_type='xdsi',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00'])
        tplot( 'erg_hep_l2_FEDU_L_energy_mag', display=display, save_png='erg_hep_energy_limit_gyro.png' )
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_energy_mag'))
        self.assertTrue('erg_hep_l2_FEDU_L_energy_mag' in vars)

    def test_hep_pad(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_hep_part_products('erg_hep_l2_FEDU_L', outputs='pa', energy=[15000., 22000.], fac_type='xdsi',
                                     mag_name=mag_vn, pos_name=pos_vn)
        tplot( 'erg_hep_l2_FEDU_L_pa', display=display, save_png='erg_hep_pad.png' )
        self.assertTrue('erg_hep_l2_FEDU_L_pa' in vars)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_pa'))

    def test_hep_en_pad_limit(self):
        del_data('*')
        # Load HEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.hep( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate energy-time spectra of electron flux for limited pitch-angle (PA) ranges
        ## Here we calculate energy-time spectra for PA = 0-10 deg and PA = 80-100 deg.
        vars1 = erg_hep_part_products('erg_hep_l2_FEDU_L', outputs='fac_energy', pitch=[80., 100.],
                                     fac_type='xdsi', mag_name=mag_vn, pos_name=pos_vn,
                                     suffix='_pa80-100')
        vars2 = erg_hep_part_products('erg_hep_l2_FEDU_L', outputs='fac_energy', pitch=[0., 10.], fac_type='xdsi',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     suffix='_pa0-10')

        ## Decorate the obtained spectrum variables
        options('erg_hep_l2_FEDU_L_energy_mag_pa80-100', 'ytitle', 'HEP-e flux\nPA: 80-100\n\n[eV]')
        options('erg_hep_l2_FEDU_L_energy_mag_pa0-10', 'ytitle', 'HEP-e flux\nPA: 0-10\n\n[eV]')
        tplot(['erg_hep_l2_FEDU_L_energy_mag_pa80-100', 'erg_hep_l2_FEDU_L_energy_mag_pa0-10'], display=display, save_png='erg_hep_en_pa_limit.png')
        self.assertTrue('erg_hep_l2_FEDU_L_energy_mag_pa0-10' in vars2)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_energy_mag_pa0-10'))
        self.assertTrue('erg_hep_l2_FEDU_L_energy_mag_pa80-100' in vars1)
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L_energy_mag_pa80-100'))

if __name__ == '__main__':
    unittest.main()


import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from matplotlib.lines import Line2D
from matplotlib.backend_bases import KeyEvent
from matplotlib.pyplot import rcParams

class TimeSelector:

    def add_selection_line(self,time):
        for item in self.axis_list:
            ax, y_bottom, y_top = item
            # Get the y-axis range
            vline = Line2D([time,time], [y_bottom, y_top], color='b', linestyle='--')
            self.selected_lines.append(vline)
            ax.add_line(vline)

    def clear_selection_lines(self):
        for vline in self.selected_lines:
            vline.set_color('r')
            vline.remove()
        plt.draw()
        self.selected_lines.clear()


    # Define the motion event handler
    def ctime_on_motion(self, event):
        if event.inaxes:  # Check if the mouse is over the plot's axes
            # Get the current x-position of the cursor (time)
            cursor_time = event.xdata
            # Update the vertical line position
            for vline in self.vertical_lines:
                vline.set_xdata([cursor_time, cursor_time])
            plt.draw()


    # Define the click event handler
    def ctime_on_click(self, event):

        if event.inaxes:  # Check if the click is within the plot's axes
            if event.button == 1:  # Left-click to add a timestamp, or shift-left-click to clear timestamps
                if self.shift_on:
                    self.clear_selection_lines()
                    self.selected_times.clear()
                    self.logs.append("shift left click")
                else:
                    # Convert the x-coordinate (time) to a datetime
                    self.logs.append("left click")
                    timestamp = mdates.num2date(event.xdata).timestamp()
                    self.selected_times.append(timestamp)
                    self.add_selection_line(event.xdata)
                plt.draw()
            elif event.button == 3:  # Right-click to stop
                self.logs.append("right click")
                plt.disconnect(self.cid)  # Disconnect the event handler
                plt.disconnect(self.motion_cid)  # Disconnect motion event handler
                plt.disconnect(self.kbd_on_cid)  # Disconnect kbd event handler
                plt.disconnect(self.kbd_off_cid)  # Disconnect kbd event handler
                for vline in self.vertical_lines:
                    vline.remove()  # Remove the vertical line from the plot
                plt.draw()
                self.saved_fig.canvas.stop_event_loop()

    # Define the key press event handler
    def ctime_on_key(self, event: KeyEvent):

        self.logs.append("key pressed: " + event.key)
        if event.key == 'c' or event.key == 'e':  # Check if 'c' was pressed
            #print("Pressed 'c', clearing selections")
            self.clear_selection_lines()
            self.selected_times.clear()
            plt.draw()
        elif event.key == 'shift':
            self.shift_on=True
        elif event.key == 'q':  # Check if 'q' was pressed
            #print("Pressed 'q', quitting...")
            plt.disconnect(self.cid)  # Disconnect the event handler
            plt.disconnect(self.motion_cid)  # Disconnect motion event handler
            plt.disconnect(self.kbd_on_cid)  # Disconnect kbd event handler
            plt.disconnect(self.kbd_off_cid)  # Disconnect kbd event handler
            for vline in self.vertical_lines:
                vline.remove()  # Remove the vertical line from the plot
            plt.draw()
            self.saved_fig.canvas.stop_event_loop()
            #clear_selection_lines()

    # Define the key release event handler
    def ctime_off_key(self, event: KeyEvent):
        self.logs.append("key released")
        if event.key == 'shift':
            self.shift_on=False

    def ctime_run_event_loop(self):
        # matplotlib defines some default keyboard commands for manipulating plots.  In particular,
        # 'c' is assigned to "navigate backward", and 'q' is assigned to "close the plot".  We want
        # to redefine these as "clear selections" and "quit ctime", then restore the original definitions
        # upon exit.
        #
        # This only works if the user hasn't redefined 'c' and 'q' to mean something else.
        # The correct way to do this would be to go through all the keymap.* keys, delete 'c' or 'q' from them,
        # then restore the whole map at the end.  This is probably good enough for now.
        #

        save_back = rcParams['keymap.back']
        save_quit = rcParams['keymap.quit']
        rcParams['keymap.back'] = []
        rcParams['keymap.quit'] = []

        #plt.draw()
        #plt.show(block=False)
        # plt.ioff()
        self.saved_fig.canvas.start_event_loop(-1)

        # Restore original keyboard shortcuts
        rcParams['keymap.back'] = save_back
        rcParams['keymap.quit'] = save_quit

        return self.selected_times

    def __init__(self, fig):
        """Initialize."""
        self.saved_fig = fig
        self.selected_times = []
        self.selected_lines = []
        self.vertical_lines = []
        self.axis_list = []
        self.cid = None
        self.motion_cid = None
        self.kbd_on_cid = None
        self.kbd_off_cid = None
        self.shift_on = False
        self.logs = []

        # Create vertical lines in each subplot to track the cursor position
        for ax in fig.axes:
            # Get the y-axis range
            y_bottom, y_top = ax.get_ylim()
            vline = Line2D([0, 0], [y_bottom, y_top], color='g', linestyle='--')
            self.vertical_lines.append(vline)
            ax.add_line(vline)
            self.axis_list.append((ax, y_bottom, y_top))

        # Connect the motion event to the handler
        self.motion_cid = fig.canvas.mpl_connect('motion_notify_event', self.ctime_on_motion)

        # Connect the click event to the handler
        self.cid = fig.canvas.mpl_connect('button_press_event', self.ctime_on_click)

        # Connect the key press/release events to the handler
        self.kbd_on_cid = fig.canvas.mpl_connect('key_press_event', self.ctime_on_key)
        self.kbd_off_cid = fig.canvas.mpl_connect('key_release_event', self.ctime_off_key)


def ctime(fig):
    """ Select time values by clicking on a plot, similar to ctime in IDL SPEDAS

    Left click saves the time at the current cursor position.  'c', 'e', or shift-left click clears the list of times selected. Right click or 'q' exits and returns the list
    of saved times (as floating point Unix times).

    Parameters
    ----------
    fig
        A matplotlib fig object specifying the plot to be used for the time picker (returned by tplot with return_plot_objects=True)

    Notes
    ------
    This feature is very platform-dependent, and should still be considered experimental.  Please let the PySPEDAS developers know
    if you have trouble using it in your environment.

    As of this release, ctime seems to work in most situations, except for Jupyter notebooks using the 'ipympl' (aka 'widget')
    matplotlib backend.  With that backend, the most common failure modes are that the ctime() call returns immediately, with no
    chance to interact with the plot, or the tplot call gets stuck somehow and never renders the plot.

    We are working on a fix for the ipympl incompatibility, but for now, the best workaround for Jupyter notebooks may be to use "%matplotlib auto"
    as the backend.  Depending on your exact environment, this will probably render the plots outside of the Jupyter notebooks (so the plot results
    won't be saved in the notebook), but at least ctime should work, allowing you to continue with your desired workflow.

    Returns
    -------
    list of double
        The selected times as floating point Unix times (in seconds)

    Examples
    --------

    >>> import pyspedas
    >>> from pyspedas import tplot, ctime, time_string
    >>> pyspedas.projects.themis.state(probe='a')
    >>> myfig, ax = tplot('tha_pos', return_plot_objects=True)
    >>> saved_timestamps = ctime(myfig)
    >>> print(time_string(saved_timestamps))

    """
    ctime_obj = TimeSelector(fig)
    print("Use the mouse to select times.  Left click to add a new time, type 'c' to clear selections, 'q' or right click to exit")
    selected_times = ctime_obj.ctime_run_event_loop()
    return selected_times




import numpy as np
from pyspedas.cotrans_tools.cart_to_sphere import cart_to_sphere
from pyspedas.cotrans_tools.sphere_to_cart import sphere_to_cart

def spd_pgs_v_shift(data, vector):
# Procedure:
# spd_pgs_v_shift
# 
# Purpose:
#   Shift a single distribution strucure by a specified velocity vector
# 
# Input:
#   data:  Sanitized particle data structure to be operated on
#   vector:  3-vector in km/s
#   matrix:  (optional) rotation matrix to apply to vector before shift
# 
# Output:
#   error:  flag, 1 indicates error, 0 none
# 
# Notes:
#   -Particle velocities are assumed to be small enough 
#    to use classical calculation.
# 

# Ensure vector has exactly 3 elements
    if len(vector) != 3:
        return

# If matrix has 9 elements, do matrix-vector multiplication
    #if matrix.size == 9:
        #vector = matrix.reshape(3, 3) @ vector

    # calculate bin velocities
    # distribution mass in eV/(km/s)^2
    v = np.sqrt(2.0 * data['energy']/data['mass'])
    theta = data['theta']
    phi = data['phi']

    #sphere to cart
    x, y, z = sphere_to_cart(v, theta, phi)

    #subtract input vector
    newx = x - vector[0]
    newy = y - vector[1]
    newz = z - vector[2]    

    #cart to sphere
    v_new, theta, phi = cart_to_sphere(newx, newy, newz)
    data['energy'] = .5 * data['mass'] * v_new**2
    data['theta'] = theta
    data['phi'] = phi



import pyspedas.projects.maven.spdf as spdf_load
from .maven_load import load_data
maven_load = load_data
from .euv import euv
from .iuv import iuv
from .kp import kp
from .lpw import lpw
from .mag import mag
from .ngi import ngi
from .rse import rse
from .sep import sep
from .sta import sta
from .swea import swea
from .swia import swia



"""Transformation SSE data to SEL data.
Notes:
    Works in a similar way to IDL spedas sse2sel.pro
"""
import logging
import numpy as np
import pyspedas
from pyspedas import tnormalize, tcrossp, data_exists, get_coords, set_coords, get_data, store_data
from pyspedas.cotrans_tools.tvector_rotate import tvector_rotate
from pyspedas.cotrans_tools.cotrans import cotrans
from pyspedas.analysis.tinterpol import tinterpol
from pyspedas.analysis.deriv_data import deriv_data
from pyspedas.projects.themis.cotrans.gse2sse import gse2sse
from pyspedas.projects.themis import autoload_support

def sse2sel(name_in: str, name_out: str, isseltosse: bool = False, ignore_input_coord: bool = False) -> int:
    """Transform sse to sel.

    Parameters
    ----------
        name_in: str
            Name of input tplot variable (e.g. 'tha_state_pos_sse' or 'tha_state_pos_sel')
        name_out: str
            Name of output tplot variable (e.g. 'tha_fgl_sse')
        isseltosse: bool
            If False (default), then SSE to SEL.
            If True, then SEL to SSE.
        ignore_input_coord: bool
            if False (default), do not check the input coordinate system
            if True, fail and return 0 if input coordinate does not match the requested transform.
        variable_type: str
            A string describing the type of data being transformed.  If value is "pos" or "vel", the appropriate
                offsets (lunar position or velocity) are applied during the transform.  Any other value will be treated
                as equivalent to rotate_only=True.
        rotate_only: bool
            if False (default), assume input variable is a position with units of km, and apply the earth-moon
                offset before rotating to SEL, or after rotating to SSE     #TODO: please check 
            if True, assume the input variable is a velocity or some other quantity that does not need the earth-moon
                translation step
    Returns
    -------
        1 for sucessful completion.
    """
    needed_vars = [name_in]
    c = [value for value in needed_vars if data_exists(value)]
    if len(c) < 1:
        logging.error("Variables needed: " + str(needed_vars))
        m = [value for value in needed_vars if value not in c]
        logging.error("Variables missing: " + str(m))
        logging.error("Please load missing variables.")
        return 0

    autoload_support(varname=name_in,slp=True)
    name_sun_pos = 'slp_sun_pos'
    name_lun_pos = 'slp_lun_pos'
    name_lun_att_x = 'slp_lun_att_x'
    name_lun_att_z = 'slp_lun_att_z'

    if not ignore_input_coord:
        # check input coord
        in_coord = get_coords(name_in)
        if in_coord is None:
            in_coord = "None"
        if not isseltosse and (in_coord.lower() != 'sse'):
            logging.error("SSE to SEL transform requested, but input coordinate system is " + in_coord)
            return 0
        if isseltosse and (in_coord.lower() != 'sel'):
            logging.error("SEL to SSE transform requested, but input coordinate system is " + in_coord)
            return 0
        
        # check sun pos coord
        sun_pos_coord = get_coords(name_sun_pos)
        if sun_pos_coord is None:
            sun_pos_coord = "None"
        if sun_pos_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE',name_sun_pos)
            sun_pos_gse_name = name_sun_pos+'_gse'
            cotrans(name_in=name_sun_pos,name_out=sun_pos_gse_name,coord_out='GSE')
        else:
            sun_pos_gse_name=name_sun_pos
        # check lun pos coord
        lun_pos_coord = get_coords(name_lun_pos)
        if lun_pos_coord is None:
            lun_pos_coord = "None"
        if lun_pos_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE', name_lun_pos)
            lun_pos_gse_name = name_lun_pos + '_gse'
            cotrans(name_in=name_lun_pos, name_out=lun_pos_gse_name, coord_out='GSE')
        else:
            lun_pos_gse_name = name_lun_pos
        # check lun att x coord
        lun_att_x_coord = get_coords(name_lun_att_x)
        if lun_att_x_coord is None:
            lun_att_x_coord = "None"
        if lun_att_x_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE', name_lun_att_x)
            lun_att_x_gse_name = name_lun_att_x + '_gse'
            cotrans(name_in=name_lun_att_x, name_out=lun_att_x_gse_name, coord_out='GSE')
        else:
            lun_att_x_gse_name = name_lun_att_x
        # check lun att z coord
        lun_att_z_coord = get_coords(name_lun_att_z)
        if lun_att_z_coord is None:
            lun_att_z_coord = "None"
        if lun_att_z_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE', name_lun_att_z)
            lun_att_z_gse_name = name_lun_att_z + '_gse'
            cotrans(name_in=name_lun_att_z, name_out=lun_att_z_gse_name, coord_out='GSE')
        else:
            lun_att_z_gse_name = name_lun_att_z
    else:
        sun_pos_gse_name = name_sun_pos
        lun_pos_gse_name = name_lun_pos
        lun_att_x_gse_name = name_lun_att_x
        lun_att_z_gse_name = name_lun_att_z
    # Make rotation matrix
    sunpos = get_data(sun_pos_gse_name)
    sun_pos_dim = sunpos.y.shape

    # X basis vector
    result = gse2sse(lun_att_x_gse_name, 'sel_x_sse', rotation_only=True)
    sel_x_sse = get_data('sel_x_sse')
    x_axis = sel_x_sse.y
    
    # Z basis vector
    result = gse2sse(lun_att_z_gse_name, 'sel_z_sse', rotation_only=True)
    sel_z_sse = get_data('sel_z_sse')
    z_axis = sel_z_sse.y

    # Y basis vector
    tcrossp('sel_z_sse', 'sel_x_sse', newname='sel_y_sse')
    sel_y_sse = get_data('sel_y_sse')
    y_axis = sel_y_sse.y

    out_data = np.zeros((sun_pos_dim[0], 3, 3))
    if not isseltosse:
        out_data[:,0,:] = x_axis
        out_data[:,1,:] = y_axis
        out_data[:,2,:] = z_axis
    else:
        # Invert sense of conversion by transposing rotation array
        out_data[:,:,0] = x_axis
        out_data[:,:,1] = y_axis
        out_data[:,:,2] = z_axis
    store_data('sel_mat_cotrans', data={'x': sunpos.times, 'y': out_data})
    if not isseltosse:
        """ SSE -> SEL
        """
        tvector_rotate('sel_mat_cotrans',name_in,newname=name_out)
        set_coords(name_out,'SEL')
        return 1
    
    else:
        """ SEL -> SSE
        """
        tvector_rotate('sel_mat_cotrans',name_in,newname=name_out)
        set_coords(name_out,'SSE')
        return 1